# -*- coding: utf-8 -*-
"""
This module collects object representing rate expressions. It is based
on the ``chemp.util._expr`` module. The API is somewhat cumbersome since
it tries to be compatible with pure python, SymPy and the underlying
units library of ChemPy (``quantities``). Consider the API to be provisional.
"""

from __future__ import (absolute_import, division, print_function)

from collections import OrderedDict
from functools import reduce
import math
from operator import add

from ..units import get_derived_unit, default_units, energy, concentration
from ..util._dimensionality import dimension_codes, base_registry
from ..util.pyutil import memoize, deprecated
from ..util._expr import Expr


_molar = getattr(default_units, 'molar', 1)  # makes module importable.


class RateExpr(Expr):
    """ Baseclass for rate expressions, see source code of e.g. MassAction & Radiolytic. """

    @classmethod
    @deprecated(use_instead=Expr.from_callback)
    def subclass_from_callback(cls, cb, cls_attrs=None):
        """ Override RateExpr.__call__

        Parameters
        ----------
        cb : callback
            With signature (variables, all_args, backend) -> scalar
            where `variables` is a dict, `all_args` a tuple and `backend` a module.
        cls_attrs : dict, optional
            Attributes to set in subclass, e.g. parameter_keys, nargs

        Examples
        --------
        >>> from chempy import Reaction
        >>> rxn = Reaction({'O2': 1, 'H2': 1}, {'H2O2': 1})  # d[H2O2]/dt = p0*exp(-p1/T)*sqrt([O2])
        >>> def cb(variables, all_args, backend):
        ...     O2, T = variables['O2'], variables['temperature']
        ...     p0, p1 = all_args
        ...     return p0*backend.sqrt(O2)*backend.exp(-p1/T)
        >>> MyRateExpr = RateExpr.subclass_from_callback(cb, dict(parameter_keys=('temperature',),nargs=2))
        >>> k = MyRateExpr([1.3e9, 4317.2])
        >>> print('%.5g' % k({'temperature': 298.15, 'O2': 1.1e-3, 'rxn': rxn}))
        22.186

        """
        class _RateExpr(cls):

            def __call__(self, variables, backend=math, **kwargs):
                return cb(variables, self.all_args(variables), backend=backend, **kwargs)
        for k, v in (cls_attrs or {}).items():
            setattr(_RateExpr, k, v)
        return _RateExpr


class RadiolyticBase(RateExpr):
    pass  # for isinstance checks


@memoize(None)
def mk_Radiolytic(*doserate_names):
    """ Create a Radiolytic rate expression

    Note that there is no mass-action dependence in the resulting
    class, i.e. the rates does not depend on any concentrations.

    Parameters
    ----------
    \*doserate_names : str instances
        Default: ('',)


    Examples
    --------
    >>> RadiolyticAlpha = mk_Radiolytic('alpha')
    >>> RadiolyticGamma = mk_Radiolytic('gamma')
    >>> dihydrogen_alpha = RadiolyticAlpha([0.8e-7])
    >>> dihydrogen_gamma = RadiolyticGamma([0.45e-7])
    >>> RadiolyticAB = mk_Radiolytic('alpha', 'beta')

    Notes
    -----
    The instance __call__ will require by default ``'density'`` and ``'doserate'``
    in variables.

    """
    if len(doserate_names) == 0:
        doserate_names = ('',)

    class _Radiolytic(RadiolyticBase):
        argument_names = tuple('radiolytic_yield{0}'.format('' if drn == '' else '_' + drn)
                               for drn in doserate_names)  # [amount/energy]
        parameter_keys = ('density',) + tuple('doserate{0}'.format('' if drn == '' else '_' + drn)
                                              for drn in doserate_names)

        def args_dimensionality(self, reaction):
            N = base_registry['amount']
            E = get_derived_unit(base_registry, 'energy')
            return (dict(zip(dimension_codes, N/E)),)*self.nargs

        def g_values(self, *args, **kwargs):
            return OrderedDict(zip(self.parameter_keys[1:], self.all_args(*args, **kwargs)))

        @deprecated(use_instead='Radiolytic.all_args')
        def g_value(self, variables, backend=math, **kwargs):
            g_val, = self.all_args(variables, backend=backend, **kwargs)
            return g_val

        def __call__(self, variables, backend=math, reaction=None, **kwargs):
            return variables['density']*reduce(add, [variables[k]*gval for k, gval in zip(
                self.parameter_keys[1:], self.all_args(variables, backend=backend, **kwargs))])

    _Radiolytic.__name__ = 'Radiolytic' if doserate_names == ('',) else ('Radiolytic_' + '_'.join(doserate_names))
    return _Radiolytic


Radiolytic = mk_Radiolytic()


class MassAction(RateExpr):
    """ Rate-expression of mass-action type

    Notes
    -----
    :meth:`__call__` requires a :class:`Reaction` instance to be passed as ``reaction``
    keyword argument.

    Examples
    --------
    >>> ma = MassAction([3.14])
    >>> from chempy import Reaction
    >>> r = Reaction.from_string('3 A -> B', param=ma)
    >>> r.rate({'A': 2}) == {'A': -75.36, 'B': 25.12}
    True

    """

    argument_names = ('rate_constant',)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return ({'time': -1, 'amount': 1-order, 'length': 3*(order - 1)},)

    def active_conc_prod(self, variables, backend=math, reaction=None):
        result = 1
        for k, v in reaction.reac.items():
            result *= variables[k]**v
        return result

    def rate_coeff(self, variables, backend=math, **kwargs):
        rat_c, = self.all_args(variables, backend=backend, **kwargs)
        return rat_c

    def __call__(self, variables, backend=math, reaction=None, **kwargs):
        return self.rate_coeff(variables, backend=backend, reaction=reaction)*self.active_conc_prod(
            variables, backend=backend, reaction=reaction, **kwargs)

    @classmethod
    def from_callback(cls, callback, attr='rate_coeff', **kwargs):
        return super(MassAction, cls).from_callback(callback, attr=attr, **kwargs)

    def string(self, *args, **kwargs):
        if self.args is None and len(self.unique_keys) == 1:
            return self.unique_keys[0]
        else:
            return super(MassAction, self).string(*args, **kwargs)

    @classmethod
    @deprecated(use_instead='MassAction.from_callback')
    def subclass_from_callback(cls, cb, cls_attrs=None):
        """ Override MassAction.__call__ """
        _RateExpr = super(MassAction, cls).subclass_from_callback(cb, cls_attrs=cls_attrs)

        def wrapper(*args, **kwargs):
            obj = _RateExpr(*args, **kwargs)
            return cls(obj)
        return wrapper


class Arrhenius(Expr):
    """ Rate expression for a Arrhenius-type of rate: c0*exp(-c1/T)

    Examples
    --------
    >>> from math import exp
    >>> from chempy import Reaction
    >>> from chempy.units import allclose, default_units as u
    >>> A = 1e11 / u.second
    >>> Ea_over_R = 42e3/8.3145 * u.K**-1
    >>> ratex = MassAction(Arrhenius([A, Ea_over_R]))
    >>> rxn = Reaction({'R'}, {'P'}, ratex)
    >>> dRdt = rxn.rate({'R': 3*u.M, 'temperature': 298.15*u.K})['R']
    >>> allclose(dRdt, -3*1e11*exp(-42e3/8.3145/298.15)*u.M/u.s)
    True

    """
    argument_names = ('A', 'Ea_over_R')
    parameter_keys = ('temperature',)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return (
            {'time': -1,
             'amount': 1-order, 'length': 3*(order - 1)},
            {'temperature': 1},
        )

    def __call__(self, variables, backend=math, **kwargs):
        A, Ea_over_R = self.all_args(variables, backend=backend, **kwargs)
        return A*backend.exp(-Ea_over_R/variables['temperature'])


class Eyring(Expr):
    """ Rate expression for Eyring eq: c0*T*exp(-c1/T)

    Note that choice of standard state (c^0) will matter for order > 1.
    """

    argument_names = ('kB_h_times_exp_dS_R', 'dH_over_R', 'conc0')
    argument_defaults = (1*_molar,)
    parameter_keys = ('temperature',)

    def args_dimensionality(self, reaction):
        order = reaction.order()
        return (
            {'time': -1, 'temperature': -1,
             'amount': 1-order, 'length': 3*(order - 1)},
            {'temperature': 1},
            concentration
        )

    def __call__(self, variables, backend=math, **kwargs):
        c0, c1, conc0 = self.all_args(variables, backend=backend, **kwargs)
        T = variables['temperature']
        return c0*T*backend.exp(-c1/T)*conc0**(1-kwargs['reaction'].order())


class EyringHS(Expr):
    argument_names = ('dH', 'dS', 'c0')
    argument_defaults = (1*_molar,)
    parameter_keys = ('temperature', 'molar_gas_constant',
                      'Boltzmann_constant', 'Planck_constant')

    def args_dimensionality(self, **kwargs):
        return (
            energy + {'amount': -1},
            energy + {'amount': -1, 'temperature': -1},
            concentration
        )

    def __call__(self, variables, backend=math, reaction=None, **kwargs):
        dH, dS, c0 = self.all_args(variables, backend=backend, **kwargs)
        T, R, kB, h = [variables[k] for k in self.parameter_keys]
        return kB/h*T*backend.exp(-(dH-T*dS)/(R*T))*c0**(1-reaction.order())


class RampedTemp(Expr):
    """ Ramped temperature, pass as substitution to e.g. ``get_odesys`` """
    argument_names = ('T0', 'dTdt')
    parameter_keys = ('time',)  # consider e.g. a parameter such as 'init_time'

    def args_dimensionality(self, **kwargs):
        return ({'temperature': 1}, {'temperature': 1, 'time': -1})

    def __call__(self, variables, backend=None, **kwargs):
        T0, dTdt = self.all_args(variables, backend=backend, **kwargs)
        return T0 + dTdt*variables['time']


class SinTemp(Expr):
    argument_names = ('Tbase', 'Tamp', 'angvel', 'phase')
    parameter_keys = ('time',)

    def args_dimensionality(self, **kwargs):
        return ({'temperature': 1}, {'temperature': 1}, {'time': -1}, {})

    def __call__(self, variables, backend=math, **kwargs):
        Tbase, Tamp, angvel, phase = self.all_args(variables, backend=backend, **kwargs)
        return Tbase + Tamp*backend.sin(angvel*variables['time'] + phase)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from collections import OrderedDict
from .pyutil import defaultnamedtuple

dimension_codes = OrderedDict(zip(
    'length mass time current temperature amount'.split(),  # not considering luminous_intensity
    'L M T I Θ N'.split()
))


class DimensionalitySI(defaultnamedtuple('DimensionalitySIBase', dimension_codes.keys(), (0,)*len(dimension_codes))):

    def __mul__(self, other):
        return self.__class__(*(x+y for x, y in zip(self, other)))

    def __truediv__(self, other):
        return self.__class__(*(x-y for x, y in zip(self, other)))

    def __pow__(self, exp):
        return self.__class__(*(x*exp for x in self))

base_registry = {name: DimensionalitySI(**{name: 1}) for name in dimension_codes}



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)


from .._util import prodpow
from ..util.testing import requires


def test_prodpow():
    result = prodpow([11, 13], [[0, 1], [1, 2]])
    assert result[0] == 13
    assert result[1] == 11*13*13


@requires('sympy')
def test_prodpow__symbols():
    import sympy
    a, b = sympy.symbols('a b')
    exprs = prodpow([a, b], [[0, 1], [1, 2]])
    assert exprs[0] == b
    assert exprs[1] == a*b**2



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from ..util.testing import requires

try:
    import numpy as np
except ImportError:
    np, NumSysLin, NumSysLog = [None]*3
else:
    from scipy.optimize import fsolve

from .._equilibrium import equilibrium_residual, solve_equilibrium, _get_rc_interval
from .._util import prodpow


@requires('numpy')
def test_equilibrium_residual():
    c0 = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14
    assert abs(equilibrium_residual(0.1, c0, stoich, K) -
               (K - (13-0.2)**-2*(11 + 0.3)**3*(17 - 0.4)**-4)) < 1e-14


@requires('numpy')
def test_get__rc_interval():
    c = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    limits = _get_rc_interval(stoich, c)
    lower = -11/3.
    upper = 17./4
    assert abs(limits[0] - lower) < 1e-14
    assert abs(limits[1] - upper) < 1e-14


@requires('numpy')
def test_solve_equilibrium_1():
    c = np.array((13., 11, 17))
    stoich = np.array((-2, 3, -4))
    K = 3.14

    def f(x):
        return (13 - 2*x)**-2 * (
            11 + 3*x)**3 * (17 - 4*x)**-4 - K
    assert np.allclose(solve_equilibrium(c, stoich, K),
                       c + stoich*fsolve(f, 3.48))


@requires('numpy')
def test_solve_equilibrium_2():
    c = np.array([1.7e-03, 3.0e+06, 3.0e+06, 9.7e+07, 5.55e+09])
    stoich = (1, 1, 0, 0, -1)
    K = 55*1e-6

    def f(x):
        return prodpow(c+x*stoich, stoich) - K
    solution = solve_equilibrium(c, stoich, K)
    assert np.allclose(solution, c + stoich*fsolve(f, 0.1))



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import math

import pytest

from chempy import Reaction, ReactionSystem, Substance
from chempy.units import allclose, Backend, to_unitless, units_library, default_units as u
from chempy.util._expr import Expr
from chempy.util.parsing import parsing_library
from chempy.util.testing import requires
from ..rates import RateExpr, MassAction, Arrhenius, Radiolytic, mk_Radiolytic, Eyring


class SpecialFraction(RateExpr):
    """
    Example from Atkins, De Paula, Physical Chemistry
    H2 + Br2 -> 2HBr
    ½ dHBr/dt = k[H2][Br2]**(3/2) / ([Br2] + k'[HBr])
    """
    def __call__(self, variables, backend=math, reaction=None):
        two = 2 * backend.pi**0
        k = self.arg(variables, 0)
        kp = self.arg(variables, 1)
        H2, Br2, HBr = variables['H2'], variables['Br2'], variables['HBr']
        return k*H2*Br2**(3/two) / (Br2 + kp*HBr)


def _get_SpecialFraction_rsys(k, kprime):
    ratex = SpecialFraction([k, kprime], ['k_HBr', 'kprime_HBr'])
    rxn = Reaction({'H2': 1, 'Br2': 1}, {'HBr': 2}, ratex)
    return ReactionSystem([rxn], 'H2 Br2 HBr')


@requires('numpy')
def test_specialfraction_rateexpr():
    rsys = _get_SpecialFraction_rsys(11, 13)
    r = rsys.rxns[0]

    conc = {'H2': 2, 'Br2': 3, 'HBr': 5}

    def _check(k, kprime, c):
        ref = k*c['H2']*c['Br2']**1.5/(c['Br2'] + kprime*c['HBr'])
        assert abs(r.rate_expr()(c) - ref) < 1e-15

    _check(11, 13, conc)


def test_RateExpr__subclass_from_callback():
    SF = RateExpr.subclass_from_callback(
        lambda v, a, backend: a[0]*v['H2']*v['Br2']**(3/2) / (v['Br2'] + a[1]*v['HBr'])
    )
    ratex = SF([11, 13], ['k_HBr', 'kprime_HBr'])
    r = Reaction({'H2': 1, 'Br2': 1}, {'HBr': 2}, ratex)

    res1 = r.rate_expr()({'H2': 5, 'Br2': 7, 'HBr': 15})
    ref1 = 11*5*7**1.5/(7+13*15)
    assert abs((res1-ref1)/ref1) < 1e-14

    res2 = r.rate_expr()({'H2': 5, 'Br2': 7, 'HBr': 15, 'k_HBr': 23, 'kprime_HBr': 42})
    ref2 = 23*5*7**1.5/(7+42*15)
    assert abs((res2-ref2)/ref2) < 1e-14


def test_MassAction():
    ma = MassAction([3.14], ['my_rate'])
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, ma, {'B': 1})
    arg1 = {'A': 11, 'B': 13, 'C': 17}
    arg2 = {'A': 11, 'B': 13, 'C': 17, 'my_rate': 2.72}
    res1 = r.rate_expr()(arg1, reaction=r)
    res2 = r.rate_expr()(arg2, reaction=r)
    ref1 = 3.14*13*11**2
    ref2 = 2.72*13*11**2
    assert abs(res1 - ref1) < 1e-14
    assert abs(res2 - ref2) < 1e-12
    rat1 = r.rate(arg1)
    rat2 = r.rate(arg2)
    for key, coeff in [('A', -2), ('B', -2), ('C', 1)]:
        assert abs(rat1[key] - ref1*coeff) < 2e-12
        assert abs(rat2[key] - ref2*coeff) < 2e-12


def test_Expr__from_callback():
    def rate_coeff(args, T, reaction, backend=math):
        return args[0]*backend.exp(args[1]/T)
    RateExpr = Expr.from_callback(rate_coeff, parameter_keys=('Tem',), nargs=2)
    k1 = RateExpr([2.1e10, -5132.2])
    rxn = Reaction({'H2': 2, 'O2': 1}, {'H2O': 2}, MassAction(k1))
    ma = rxn.rate_expr()
    res = ma({'Tem': 273.15, 'H2': 7, 'O2': 13}, reaction=rxn)
    ref = 7*7*13*2.1e10*math.exp(-5132.2/273.15)
    assert abs((res-ref)/ref) < 1e-14


def test_MassAction__subclass_from_callback():
    def rate_coeff(variables, all_args, reaction, backend):
        return all_args[0]*backend.exp(all_args[1]/variables['temperature'])
    CustomMassAction = MassAction.subclass_from_callback(
        rate_coeff, cls_attrs=dict(parameter_keys=('temperature',), nargs=2))
    k1 = CustomMassAction([2.1e10, -5132.2])
    rxn = Reaction({'H2': 2, 'O2': 1}, {'H2O': 2}, k1)
    cma = rxn.rate_expr()
    res = cma({'temperature': 273.15, 'H2': 7, 'O2': 13}, reaction=rxn)
    ref = 7*7*13*2.1e10*math.exp(-5132.2/273.15)
    assert abs((res-ref)/ref) < 1e-14


@requires(units_library)
def test_MassAction__subclass_from_callback__units():
    def rate_coeff(variables, all_args, backend, **kwargs):
        return all_args[0]*backend.exp(all_args[1]/variables['temperature'])
    CustomMassAction = MassAction.subclass_from_callback(
        rate_coeff, cls_attrs=dict(parameter_keys=('temperature',), nargs=2))
    k1 = CustomMassAction([2.1e10/u.molar**2/u.second, -5132.2*u.kelvin])
    rxn = Reaction({'H2': 2, 'O2': 1}, {'H2O': 2}, k1)
    variables = {
        'temperature': 491.67*u.rankine,
        'H2': 7000*u.mol/u.metre**3,
        'O2': 13*u.molar
    }
    cma = rxn.rate_expr()
    res = cma(variables, backend=Backend(), reaction=rxn)
    ref = 7*7*13*2.1e10*math.exp(-5132.2/273.15) * u.molar/u.second
    assert allclose(res, ref)


def test_ArrheniusMassAction():
    A, Ea_over_R = 1.2e11, 40e3/8.3145
    ama = MassAction(Arrhenius([A, Ea_over_R]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, ama, {'B': 1})
    T_ = 'temperature'

    def ref(v):
        return 1.2e11*math.exp(-Ea_over_R/v[T_])*v['B']*v['A']**2

    ma = r.rate_expr()
    for params in [(11., 13., 17., 311.2),
                   (12, 8, 5, 270)]:
        var = dict(zip(['A', 'B', 'C', T_], params))
        ref_var = ref(var)
        assert abs((ma(var, reaction=r) - ref_var)/ref_var) < 1e-14

    with pytest.raises(ValueError):
        Arrhenius([A, Ea_over_R, 1, A])

    # assert ama.as_mass_action({T_: 273.15}).args[0] == A*math.exp(-Ea_over_R/273.15)


@requires(units_library)
def test_specialfraction_rateexpr__units():
    _k, _kprime = 3.5 * u.s**-1 * u.molar**-0.5, 9.2
    rsys = _get_SpecialFraction_rsys(_k, _kprime)
    r = rsys.rxns[0]

    conc = {'H2': 2*u.molar, 'Br2': 3000*u.mol/u.m**3, 'HBr': 5*u.molar}
    ma = r.rate_expr()

    def _check(k, kprime, c):
        ref = k*c['H2']*c['Br2']**1.5/(c['Br2'] + kprime*c['HBr'])
        assert abs(ma(c, reaction=r) - ref) < 1e-15*u.molar/u.second

    _check(_k, _kprime, conc)
    alt_k, alt_kprime = 2*u.s**-1*u.molar**-0.5, 5
    _check(alt_k, alt_kprime, dict(list(conc.items()) + [
        ('k_HBr', alt_k), ('kprime_HBr', alt_kprime)]))


@requires(units_library)
def test_ArrheniusMassAction__units():
    A, Ea_over_R = 1.2e11/u.molar**2/u.second, 40e3/8.3145*u.kelvin
    ama = MassAction(Arrhenius([A, Ea_over_R]))
    r = Reaction({'A': 2, 'B': 1}, {'C': 1}, ama, {'B': 1})
    T_ = 'temperature'

    def ref(v):
        return 1.2e11/u.molar**2/u.second*math.exp(
            -Ea_over_R/v[T_])*v['B']*v['A']**2
    ma = r.rate_expr()

    for params in [(11.*u.molar, 13.*u.molar, 17.*u.molar, 311.2*u.kelvin),
                   (12*u.molar, 8*u.molar, 5*u.molar, 270*u.kelvin)]:
        var = dict(zip(['A', 'B', 'C', T_], params))
        ref_val = ref(var)
        assert abs((ma(var, reaction=r) - ref_val)/ref_val) < 1e-14


def test_Radiolytic():
    r = Radiolytic([2.1e-7])
    res = r({'doserate': 0.15, 'density': 0.998})
    assert abs(res - 0.15*0.998*2.1e-7) < 1e-15


@requires(parsing_library)
def test_Radiolytic__parsing():
    rxn = Reaction.from_string("-> H + OH; Radiolytic({'radiolytic_yield': 2.1e-7})", None)
    res = rxn.rate({'doserate': 0.15, 'density': 0.998})
    ref = 0.15*0.998*2.1e-7
    assert abs((res['H'] - ref)/ref) < 1e-15
    assert abs((res['OH'] - ref)/ref) < 1e-15
    gval, = rxn.rate_expr().g_values({}).values()
    assert abs(gval - 2.1e-7) < 1e-15


@requires(parsing_library, units_library)
def test_Radiolytic__parsing__units():
    rxn = Reaction.from_string("-> H + OH; Radiolytic({'radiolytic_yield': 2.1e-7*mol/J})", None)
    assert rxn.reac == {}
    assert rxn.prod == {'H': 1, 'OH': 1}
    res = rxn.rate({'doserate': 0.15*u.gray/u.s, 'density': 0.998*u.kg/u.dm3})
    ref = 0.15*0.998*2.1e-7*u.molar/u.second
    assert abs((res['H'] - ref)/ref) < 1e-15
    assert abs((res['OH'] - ref)/ref) < 1e-15


@requires(units_library)
def test_Radiolytic__units():

    def _check(r):
        res = r({'doserate': 0.15*u.gray/u.second,
                 'density': 0.998*u.kg/u.decimetre**3})
        ref = 0.15*0.998*2.1e-7*u.molar/u.second
        assert abs(to_unitless((res - ref)/ref)) < 1e-15

    _check(Radiolytic([2.1e-7*u.mol/u.joule]))
    _check(Radiolytic([2.0261921896167396*u.per100eV]))


@requires(units_library)
def test_Radioyltic__Reaction_html():
    rate = Radiolytic([2.1*u.per100eV])
    rxn = Reaction({}, {'H': 1}, rate)
    H = Substance.from_formula('H')
    html = rxn.html({'H': H}, with_param=True)
    assert html == ' &rarr; H&#59; %s' % str(rate)


def test_mk_Radiolytic():
    R1 = mk_Radiolytic()
    R2 = mk_Radiolytic()
    assert R1 is R2

    RABG = mk_Radiolytic('alpha', 'beta', 'gamma')
    rxn = Reaction({}, {'H': 2}, RABG([3, 5, 7], 'ya yb yg'.split()))
    rat = rxn.rate({'doserate_alpha': 11, 'doserate_beta': 13, 'doserate_gamma': 17, 'density': .7})
    assert abs(rat['H'] - .7*2*(3*11 + 5*13 + 7*17)) < 1e-13
    assert RABG.parameter_keys == ('density', 'doserate_alpha', 'doserate_beta', 'doserate_gamma')
    assert RABG.argument_names == tuple('radiolytic_yield_%s' % k for k in 'alpha beta gamma'.split())
    assert rxn.param.unique_keys == ('ya', 'yb', 'yg')
    rat2 = rxn.rate({'doserate_alpha': 11, 'doserate_beta': 13, 'doserate_gamma': 17, 'density': .7,
                     'ya': 23, 'yb': 29, 'yg': 31})
    assert abs(rat2['H'] - .7*2*(23*11 + 29*13 + 31*17)) < 1e-13


class Eyring4(Expr):
    nargs = 4

    def __call__(self, variables, reaction, backend=math):
        Sact_fact, Hact_exp, Sref_fact, Href_exp = self.all_args(variables, backend=backend)
        T = variables['temperature']
        return T * Sact_fact / Sref_fact * backend.exp((Href_exp-Hact_exp)/T)


def test_EyringMassAction():
    args = kB_h_times_exp_dS_R, dH_over_R, c0 = 1.2e11/273.15, 40e3/8, 1
    ama = MassAction(Eyring(args, ('Sfreq', 'Hact')))
    rxn1 = Reaction({'A': 2, 'B': 1}, {'C': 1}, ama, {'B': 1})
    T_ = 'temperature'

    def ref(v):
        return v.get('Sfreq', 1.2e11/273.15)*v[T_]*math.exp(-v.get('Hact', 40e3/8)/v[T_])*v['B']*v['A']**2

    for params in [(11., 13., 17., 311.2),
                   (12, 8, 5, 270)]:
        var = dict(zip(['A', 'B', 'C', T_], params))
        ref_val = ref(var)
        assert abs((ama(var, reaction=rxn1) - ref_val)/ref_val) < 1e-14

    with pytest.raises(ValueError):
        MassAction(Eyring([1, 1, 1, 1, 1]))

    # assert ama.as_mass_action({T_: 273.15}).args[0] == 1.2e11*math.exp(-40e3/8/273.15)

    ama2 = MassAction(Eyring4([1.2e11/273, 40e3/8, 1.2, 1e3], ('Sfreq', 'Hact', 'Sref', 'Href')))
    rxn2 = Reaction({'C': 1}, {'A': 2, 'B': 2}, ama2)
    var2 = {'C': 29, 'temperature': 273}

    def ref2(var):
        return var['C']*var.get('temperature', 273)*var.get('Sfreq', 1.2e11/273)/var.get('Sref', 1.2)*math.exp(
            (var.get('Href', 1e3) - var.get('Hact', 5e3))/var.get('temperature', 273))

    r2 = ref2(var2)
    assert abs((ama2(var2, reaction=rxn2) - r2)/r2) < 1e-14

    rsys = ReactionSystem([rxn1, rxn2])
    var3 = {'A': 11, 'B': 13, 'C': 17, 'temperature': 298, 'Sfreq': 1.2e11/298}
    rates = rsys.rates(var3)
    rf3 = ref(var3)
    rb3 = ref2(var3)
    ref_rates = {'A': 2*(rb3 - rf3), 'B': 2*(rb3 - rf3), 'C': rf3 - rb3}
    for k, v in ref_rates.items():
        assert abs((rates[k] - v)/v) < 1e-14



# -*- coding: utf-8 -*-
"""
Convenince functions for representing reaction systems as graphs.
"""
import os
import subprocess
import shutil
import tempfile


def rsys2dot(rsys, tex=False, rprefix='r', rref0=1, nodeparams='[label="{}",shape=diamond]',
             colors=('maroon', 'darkgreen'), penwidths=None, include_inactive=True):
    """
    Returns list of lines of DOT (graph description language)
    formated graph.

    Parameters
    ==========
    rsys: ReactionSystem
    tex: bool (default False)
        If set True, output will be LaTeX formated
    (Substance need to have latex_name attribute set)
    rprefix: string
        Reaction enumeration prefix, default: r
    rref0: integer
        Reaction enumeration inital counter value, default: 1
    nodeparams: string
        DOT formated param list, default: [label={} shape=diamond]

    Returns
    =======
    list of lines of DOT representation of the graph representation.

    """
    lines = ['digraph "' + str(rsys.name) + '" {\n']
    ind = '  '  # indentation
    if penwidths is None:
        penwidths = [1.0]*rsys.nr

    categories = rsys.categorize_substances(checks=())

    def add_substance(key):
        fc = 'black'
        if key in categories['depleted']:
            fc = colors[0]
        if key in categories['accumulated']:
            fc = colors[1]
        label = ('$%s$' if tex else '%s') % getattr(rsys.substances[key], 'latex_name' if tex else 'name')
        lines.append(ind + '"{key}" [fontcolor={fc} label="{lbl}"];\n'.format(key=key, fc=fc, lbl=label))

    for sk in rsys.substances:
        add_substance(sk)

    def add_vertex(key, num, reac, penwidth):
        snum = str(num) if num > 1 else ''
        fmt = ','.join(
            ['label="{}"'.format(snum)] +
            (['penwidth={}'.format(penwidth)] if penwidth != 1 else [])
        )
        lines.append(ind + '"{}" -> "{}" [color={},fontcolor={},{}];\n'.format(
            *((key, rid, colors[0], colors[0], fmt) if reac else
              (rid, key, colors[1], colors[1], fmt))
        ))

    if include_inactive:
        reac_stoichs = rsys.all_reac_stoichs()
        prod_stoichs = rsys.all_prod_stoichs()
    else:
        reac_stoichs = rsys.active_reac_stoichs()
        prod_stoichs = rsys.active_prod_stoichs()

    for ri, rxn in enumerate(rsys.rxns):
        rid = rprefix + str(ri+rref0)
        lines.append(ind + '{')
        lines.append(ind*2 + 'node ' + nodeparams.format(rxn.name or rid))
        lines.append(ind*2 + rid)
        lines.append(ind + '}\n')
        for idx, key in enumerate(rsys.substances):
            num = reac_stoichs[ri, idx]
            if num == 0:
                continue
            add_vertex(key, num, True, penwidths[ri])
        for idx, key in enumerate(rsys.substances):
            num = prod_stoichs[ri, idx]
            if num == 0:
                continue
            add_vertex(key, num, False, penwidths[ri])
    lines.append('}\n')
    return lines


def rsys2graph(rsys, fname, output_dir=None, prog=None, save=False, **kwargs):
    """
    Convenience function to call `rsys2dot` and write output to file
    and render the graph

    Parameters
    ----------
    rsys : ReactionSystem
    fname : str
        filename
    output_dir : str (optional)
        path to directory (default: temporary directory)
    prog : str (optional)
        default: 'dot'
    save : bool
        removes temporary directory if False, default: False
    \\*\\*kwargs :
        Keyword arguments passed along to py:func:`rsys2dot`.

    Returns
    -------
    str
        Outpath

    Examples
    --------
    >>> rsys2graph(rsys, sbstncs, '/tmp/out.png')  # doctest: +SKIP

    """

    lines = rsys2dot(rsys, **kwargs)
    created_tempdir = False
    try:
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            created_tempdir = True
        basename, ext = os.path.splitext(os.path.basename(fname))
        outpath = os.path.join(output_dir, fname)
        dotpath = os.path.join(output_dir, basename + '.dot')
        with open(dotpath, 'wt') as ofh:
            ofh.writelines(lines)
        if ext == '.tex':
            cmds = [prog or 'dot2tex']
        else:
            cmds = [prog or 'dot', '-T'+outpath.split('.')[-1]]
        p = subprocess.Popen(cmds + [dotpath, '-o', outpath])
        retcode = p.wait()
        if retcode:
            fmtstr = "{}\n returned with exit status {}"
            raise RuntimeError(fmtstr.format(' '.join(cmds), retcode))
        return outpath
    finally:
        if save is True or save == 'True':
            pass
        else:
            if save is False or save == 'False':
                if created_tempdir:
                    shutil.rmtree(output_dir)
            else:
                # interpret save as path to copy pdf to.
                shutil.copy(outpath, save)




class _RxnTable(object):

    _rsys_meth = None

    def __init__(self, idx_rxn_pairs, substances, colors=None, missing=None, missing_color='eee8aa'):
        self.idx_rxn_pairs = idx_rxn_pairs
        self.substances = substances
        self.colors = colors or {}
        self.missing = missing or []
        self.missing_color = missing_color

    @classmethod
    def from_ReactionSystem(cls, rsys, color_categories=True):
        idx_rxn_pairs, unconsidered_ri = getattr(rsys, cls._rsys_meth)()
        colors = rsys._category_colors() if color_categories else {}
        missing = [not rsys.substance_participation(sk) for sk in rsys.substances]
        return cls(idx_rxn_pairs, rsys.substances, colors=colors, missing=missing), unconsidered_ri

    def _repr_html_(self):
        from .web import css
        return css(self, substances=self.substances, colors=self.colors)

    def _cell_label_html(self, ori_idx, rxn):
        """ Reaction formatting callback. (reaction index -> string) """
        pretty = rxn.unicode(self.substances, with_param=True)
        return '<a title="%d: %s">%s</a>' % (ori_idx, pretty, rxn.name or rxn.param)

    def _cell_html(self, A, ri, ci=None):
        args = []
        if ci is not None and ri > ci:
            r = '-'
        else:
            if ci is None:  # A is a vector
                c = A[ri]
                is_missing = self.missing[ri]
            else:  # A is a matrix
                c = A[ri][ci]
                is_missing = self.missing[ri] or self.missing[ci]

            if c is None:
                r = ''
            else:
                r = ', '.join(self._cell_label_html(*r) for r in c)

            if is_missing:
                args.append('style="background-color: #%s;"' % self.missing_color)

        return '<td %s>%s</td>' % (' '.join(args), r)


class UnimolecularTable(_RxnTable):
    """ Table of unimolecular reactions in a ReactionSystem

    Parameters
    ----------
    rsys : ReactionSystem
    sinks_sources_disjoint : tuple, None or True
        Colors sinks & sources. When ``True`` :meth:`sinks_sources_disjoint` is called.
    html_cell_label : Reaction formatting callback
        The function takes an integer, a Reaction instance and a dict of Substances as
        parameters and return a string.

    Returns
    -------
    string: html representation
    list: reactions not considered
    """

    _rsys_meth = '_unimolecular_reactions'

    def _html(self, printer, **kwargs):
        if 'substances' not in kwargs:
            kwargs['substances'] = self.substances
        ss = printer._get('substances', **kwargs)
        rows = '\n'.join('<tr><td>%s</td>%s</tr>' % (
            printer._print(s), self._cell_html(self.idx_rxn_pairs, rowi)
        ) for rowi, s in enumerate(ss.values()))
        return '<table>%s</table>' % rows


class BimolecularTable(_RxnTable):
    """ Table of bimolecular reactions

    Parameters
    ----------
    idx_rxn_pairs : iterable of (int, Reaction) pairs
    substances : dict
        Mapping substance key to Substance instance.
    sinks_sources_disjoint : tuple, None or True
        Colors sinks & sources. When ``True`` :meth:`sinks_sources_disjoint` is called.

    Returns
    -------
    string: html representation
    list: reactions not considered
    """

    _rsys_meth = '_bimolecular_reactions'

    def _html(self, printer, **kwargs):
        if 'substances' not in kwargs:
            kwargs['substances'] = self.substances
        ss = printer._get('substances', **kwargs)
        header = '<th></th>' + ''.join('<th>%s</th>' % printer._print(s) for
                                       s in ss.values())
        rows = ['<tr><td>%s</td>%s</tr>' % (
            printer._print(s), ''.join(self._cell_html(self.idx_rxn_pairs, rowi, ci)
                                       for ci in range(len(ss)))
        ) for rowi, s in enumerate(ss.values())]
        return '<table>%s</table>' % '\n'.join([header, '\n'.join(rows)])



# -*- coding: utf-8 -*-
"""
Contains functions for the `Eyring equation
<https://en.wikipedia.org/wiki/Eyring_equation>`_.
"""
from __future__ import (absolute_import, division, print_function)

import math

from .._util import get_backend
from ..util.regression import least_squares
from ..util.pyutil import defaultnamedtuple
from ..units import default_units, Backend, default_constants, format_string
from .arrhenius import _get_R, _fit

try:
    import numpy as np
except ImportError:
    np = None


def _get_kB_over_h(constants=None, units=None):
    if constants is None:
        kB_over_h = 2.083664399411865234375e10
        if units is not None:
            s = units.second
            K = units.kelvin
            kB_over_h /= s*K
    else:
        kB_over_h = (constants.Boltzmann_constant /
                     constants.Planck_constant)
    return kB_over_h


def eyring_equation(dH, dS, T, constants=None, units=None, backend=None):
    """
    Returns the rate coefficient according to the Eyring equation

    Parameters
    ----------
    dH: float with unit
        Enthalpy of activation.
    dS: float with unit
        Entropy of activation.
    T: float with unit
        temperature
    constants: object (optional, default: None)
        if None:
            T assumed to be in Kelvin, Ea in J/(K mol)
        else:
            attributes accessed: molar_gas_constant
            Tip: pass quantities.constants
    units: object (optional, default: None)
        attributes accessed: Joule, Kelvin and mol
    backend: module (optional)
        module with "exp", default: numpy, math

    """
    be = get_backend(backend)
    R = _get_R(constants, units)
    kB_over_h = _get_kB_over_h(constants, units)

    try:
        RT = (R*T).rescale(dH.dimensionality)
    except AttributeError:
        RT = R*T

    try:
        kB_over_h = kB_over_h.simplified
    except AttributeError:
        pass

    return kB_over_h*T*be.exp(dS/R)*be.exp(-dH/RT)


def fit_eyring_equation(T, k, kerr=None, linearized=False, constants=None, units=None):
    """ Curve fitting of the Eyring equation to data points

    Parameters
    ----------
    T : float
    k : array_like
    kerr : array_like (optional)
    linearized : bool

    """
    R = _get_R(constants, units)
    ln_kb_over_h = math.log(_get_kB_over_h(constants, units))
    return _fit(T, k, kerr, eyring_equation, lambda T, k: 1/T, lambda T, k: np.log(k/T),
                [lambda p: -p[1]*R, lambda p: R*(p[0] - ln_kb_over_h)], linearized=linearized)


class EyringParam(defaultnamedtuple('EyringParam', 'dH dS ref', [None])):
    """ Kinetic data in the form of an Eyring parameterisation

    Parameters
    ----------
    dH : float
        Enthalpy of activation.
    dS : float
        Entropy of activation.
    ref: object (default: None)
        arbitrary reference (e.g. citation key or dict with bibtex entries)

    Examples
    --------
    >>> k = EyringParam(72e3, 61.4)
    >>> '%.5g' % k(298.15)
    '2435.4'

    """

    def __call__(self, T, constants=None, units=None, backend=None):
        """ Evaluates the Eyring equation for a specified state

        Parameters
        ----------
        T : float
        constants : module (optional)
        units : module (optional)
        backend : module (default: math)

        See also
        --------
        See :func:`chempy.eyring.eyring_equation`.

        """
        return eyring_equation(self.dH, self.dS, T, constants=constants,
                               units=units, backend=backend)

    def kB_h_times_exp_dS_R(self, constants=None, units=None, backend=math):
        R = _get_R(constants, units)
        kB_over_h = _get_kB_over_h(constants, units)
        return kB_over_h * backend.exp(self.dS/R)

    def dH_over_R(self, constants=None, units=None, backend=None):
        R = _get_R(constants, units)
        return self.dH/R

    def as_RateExpr(self, unique_keys=None, constants=None, units=None, backend=math):
        from .rates import Eyring, MassAction
        args = [self.kB_h_times_exp_dS_R(constants, units, backend),
                self.dH_over_R(constants, units)]
        return MassAction(Eyring(args, unique_keys))

    def format(self, precision, tex=False):
        try:
            str_A, str_A_unit = format_string(self.A, precision, tex)
            str_Ea, str_Ea_unit = format_string(self.Ea, precision, tex)
        except Exception:
            str_A, str_A_unit = precision.format(self.A), '-'
            str_Ea, str_Ea_unit = precision.format(self.Ea), '-'
        return (str_A, str_A_unit), (str_Ea, str_Ea_unit)

    def equation_as_string(self, precision, tex=False):
        (str_A, str_A_unit), (str_Ea, str_Ea_unit) = self.format(precision, tex)
        if tex:
            return (
                r"\frac{{k_B T}}{{h}}\exp \left(\frac{{{}}}{{R}} \right)"
                r" \exp \left(-\frac{{{}}}{{RT}} \right)").format(
                    str_A, str_Ea + ' ' + str_Ea_unit), str_A_unit
        else:
            return "kB*T/h*exp({}/R)*exp(-{}/(R*T))".format(
                str_A, str_Ea + ' ' + str_Ea_unit), str_A_unit

    def __str__(self):
        return self.equation_as_string('%.5g')


class EyringParamWithUnits(EyringParam):
    def __call__(self, state, constants=default_constants, units=default_units,
                 backend=None):
        """ See :func:`chempy.eyring.eyring_equation`. """
        if backend is None:
            backend = Backend()
        return super(EyringParamWithUnits, self).__call__(
            state, constants, units, backend)

    def as_RateExpr(self, unique_keys=None, constants=default_constants,
                    units=default_units, backend=None):
        if backend is None:
            backend = Backend()
        return super(EyringParamWithUnits, self).as_RateExpr(
            unique_keys, constants, units, backend)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)


import os
import subprocess
import shutil
import tempfile

import pytest

from chempy import Reaction, ReactionSystem, Substance
from ..graph import rsys2dot, rsys2graph
from ..testing import requires

try:
    dot_missing = subprocess.call(['dot', '-?']) != 0
except OSError:
    dot_missing = True


def _get_rsys():
    r1 = Reaction({'A': 2}, {'B': 1}, param=3.0)
    A = Substance('A', latex_name='\\boldsymbol{A}')
    B = Substance('B', latex_name='\\boldsymbol{B}')
    rsys = ReactionSystem([r1], [A, B])
    return rsys


@requires('numpy')
@pytest.mark.skipif(dot_missing, reason='graphviz not installed? (dot command missing)')
def test_rsys2dot():
    rsys = _get_rsys()
    assert list(map(str.strip, rsys2dot(rsys))) == [
        'digraph "None" {',
        '"A" [fontcolor=maroon label="A"];',
        '"B" [fontcolor=darkgreen label="B"];',
        '{',
        'node [label="r1",shape=diamond]',
        'r1',
        '}',
        '"A" -> "r1" [color=maroon,fontcolor=maroon,label="2"];',
        '"r1" -> "B" [color=darkgreen,fontcolor=darkgreen,label=""];',
        '}'
    ]


@pytest.mark.slow
@requires('numpy')
@pytest.mark.skipif(dot_missing, reason='graphviz not installed? (dot command missing)')
def test_rsys2graph():
    rsys = _get_rsys()
    tempdir = tempfile.mkdtemp()
    try:
        rsys2graph(rsys, os.path.join(tempdir, 'out.png'))
        rsys2graph(rsys, os.path.join(tempdir, 'out.ps'))
        try:
            subprocess.call(['dot2tex', '-v'])
        except Exception:
            pass
        else:
            rsys2graph(rsys, os.path.join(tempdir, 'out.tex'))
    finally:
        shutil.rmtree(tempdir)



# -*- coding: utf-8 -*-
"""
This package collects utility functions used throughout the ``ChemPy`` package.
"""
from __future__ import (absolute_import, division, print_function)

from .pyutil import NoConvergence



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from ..pyutil import defaultkeydict, defaultnamedtuple


def test_defaultnamedtuple():
    Point2 = defaultnamedtuple('Point2', 'x y', [10])
    p = Point2(3)
    assert p.x == 3 and p.y == 10

    Point3 = defaultnamedtuple('Point3', 'x y z', [10, 20])
    p = Point3(3)
    assert p.x == 3 and p.y == 10 and p.z == 20

    p = Point3(3, z=30)
    assert p.x == 3 and p.y == 10 and p.z == 30

    p = Point3(3, 4, 5)
    assert p.x == 3 and p.y == 4 and p.z == 5

    _default_y = Point2.__new__.__defaults__[-1]

    class MySubclass(Point2):

        def __new__(cls, x2, y2=_default_y):
            return super(MySubclass, cls).__new__(cls, x2**.5, y2**.5)

    p2 = MySubclass(9, 4)
    assert isinstance(p2, tuple)
    assert isinstance(p2, Point2)
    assert isinstance(p2, MySubclass)
    assert not isinstance(p, MySubclass)
    assert p2.x == 3
    assert p2.y == 2

    p3 = MySubclass(9)
    assert p3.x == 3
    assert p3.y == 10**.5


def test_defaultkeydict():
    d = defaultkeydict(lambda k: k*2)
    assert d['as'] == 'asas'



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from collections import defaultdict, OrderedDict
from itertools import permutations
import math

import pytest
try:
    import numpy as np
except ImportError:
    np = None


from chempy import Equilibrium, Reaction, ReactionSystem, Substance
from chempy.thermodynamics.expressions import MassActionEq
from chempy.units import (
    SI_base_registry, get_derived_unit, allclose, units_library,
    to_unitless, default_constants as const, default_units as u
)
from chempy.util._expr import Expr
from chempy.util.testing import requires
from .test_rates import _get_SpecialFraction_rsys
from ..arrhenius import ArrheniusParam
from ..rates import Arrhenius, MassAction, Radiolytic, RampedTemp
from .._rates import ShiftedTPoly
from ..ode import get_odesys, chained_parameter_variation, _create_odesys as create_odesys
from ..integrated import dimerization_irrev, binary_rev


@requires('numpy', 'pyodesys')
def test_get_odesys_1():
    k = .2
    a = Substance('A')
    b = Substance('B')
    r = Reaction({'A': 1}, {'B': 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ['A', 'B']
    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {
        'A': 1.0,
        'B': 3.0,
    }
    t = np.linspace(0.0, 10.0)
    xout, yout, info = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k*t)
    yref[:, 1] = 4 - np.exp(-k*t)
    assert np.allclose(yout, yref)


@requires('numpy', 'pyodesys', 'sympy')
def test_get_odesys__rate_exprs_cb():
    k = .2
    a = Substance('A')
    b = Substance('B')
    r = Reaction({'A': 1}, {'B': 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ['A', 'B']
    odesys, extra = get_odesys(rsys)
    c0 = {'A': 1.0, 'B': 3.0}
    t = np.linspace(0.0, 10.0)
    res = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k*t)
    yref[:, 1] = 4 - np.exp(-k*t)
    assert np.allclose(res.yout, yref)
    rate = extra['rate_exprs_cb'](res.xout, res.yout, res.params)
    assert np.allclose(rate[:, 0], k*yref[:, 0])


@requires('numpy', 'pyodesys')
def test_get_odesys_2():
    g = Radiolytic([3.14])
    a = Substance('A')
    b = Substance('B')
    r = Reaction({'A': 1}, {'B': 1}, param=g)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {
        'A': 1.0,
        'B': 3.0,
    }
    t = np.linspace(0.0, .1)
    xout, yout, info = odesys.integrate(t, rsys.as_per_substance_array(c0),
                                        {'doserate': 2.72, 'density': .998})
    yref = np.zeros((t.size, 2))
    k = 3.14*2.72*.998
    yref[:, 0] = 1 - k*t
    yref[:, 1] = 3 + k*t
    assert np.allclose(yout, yref)


@requires(units_library, 'pyodesys')
def test_get_odesys_3():
    M = u.molar
    s = u.second
    mol = u.mol
    m = u.metre
    substances = list(map(Substance, 'H2O H+ OH-'.split()))
    dissociation = Reaction({'H2O': 1}, {'H+': 1, 'OH-': 1}, 2.47e-5/s)
    recombination = Reaction({'H+': 1, 'OH-': 1}, {'H2O': 1}, 1.37e11/M/s)
    rsys = ReactionSystem([dissociation, recombination], substances)
    odesys = get_odesys(
        rsys, include_params=True, unit_registry=SI_base_registry,
        output_conc_unit=M)[0]
    c0 = {'H2O': 55.4*M, 'H+': 1e-7*M, 'OH-': 1e-4*mol/m**3}
    x, y, p = odesys.to_arrays(-42*u.second,
                               rsys.as_per_substance_array(c0, unit=M), ())
    fout = odesys.f_cb(x, y, p)

    time_unit = get_derived_unit(SI_base_registry, 'time')
    conc_unit = get_derived_unit(SI_base_registry, 'concentration')

    r1 = to_unitless(55.4*2.47e-5*M/s, conc_unit/time_unit)
    r2 = to_unitless(1e-14*1.37e11*M/s, conc_unit/time_unit)
    assert np.all(abs(fout[:, 0] - r2 + r1)) < 1e-10
    assert np.all(abs(fout[:, 1] - r1 + r2)) < 1e-10
    assert np.all(abs(fout[:, 2] - r1 + r2)) < 1e-10


@requires(units_library, 'pyodesys')
def test_get_odesys__with_units():
    a = Substance('A')
    b = Substance('B')
    molar = u.molar
    second = u.second
    r = Reaction({'A': 2}, {'B': 1}, param=1e-3/molar/second)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True,
                        unit_registry=SI_base_registry)[0]
    c0 = {
        'A': 13*u.mol / u.metre**3,
        'B': .2 * u.molar
    }
    conc_unit = get_derived_unit(SI_base_registry, 'concentration')
    t = np.linspace(0, 10)*u.hour
    xout, yout, info = odesys.integrate(
        t, rsys.as_per_substance_array(c0, unit=conc_unit),
        atol=1e-10, rtol=1e-12)

    t_unitless = to_unitless(xout, u.second)
    Aref = dimerization_irrev(t_unitless, 1e-6, 13.0)
    # Aref = 1/(1/13 + 2*1e-6*t_unitless)
    yref = np.zeros((xout.size, 2))
    yref[:, 0] = Aref
    yref[:, 1] = 200 + (13-Aref)/2
    assert allclose(yout, yref*conc_unit)


@requires('numpy', 'pyodesys')
def test_SpecialFraction():
    k, kprime = 3.142, 2.718
    rsys = _get_SpecialFraction_rsys(k, kprime)

    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {'H2': 13, 'Br2': 17, 'HBr': 19}
    r = k*c0['H2']*c0['Br2']**(3/2)/(c0['Br2'] + kprime*c0['HBr'])
    ref = rsys.as_per_substance_array({'H2': -r, 'Br2': -r, 'HBr': 2*r})
    res = odesys.f_cb(0, rsys.as_per_substance_array(c0))
    assert np.allclose(res, ref)


@requires(units_library, 'pyodesys')
def test_SpecialFraction_with_units():
    k, kprime = 3.142 * u.s**-1 * u.molar**-0.5, 2.718
    rsys = _get_SpecialFraction_rsys(k, kprime)
    odesys = get_odesys(rsys, include_params=True,
                        unit_registry=SI_base_registry)[0]
    c0 = {'H2': 13*u.molar, 'Br2': 16*u.molar, 'HBr': 19*u.molar}
    r = k*c0['H2']*c0['Br2']**(3/2)/(c0['Br2'] + kprime*c0['HBr'])
    conc_unit = u.mol/u.metre**3
    rate_unit = conc_unit/u.second
    ref = rsys.as_per_substance_array({'H2': -r, 'Br2': -r, 'HBr': 2*r}, unit=rate_unit)
    res = odesys.f_cb(0, rsys.as_per_substance_array(c0, unit=conc_unit))
    assert allclose(to_unitless(ref, rate_unit), res)


@requires('pyodesys')
def test_ode_with_global_parameters():
    ratex = MassAction(Arrhenius([1e10, 40e3/8.3145]))
    rxn = Reaction({'A': 1}, {'B': 1}, ratex)
    rsys = ReactionSystem([rxn], 'A B')
    odesys, extra = get_odesys(rsys, include_params=False)
    param_keys, unique_keys, p_units = map(extra.get, 'param_keys unique p_units'.split())
    conc = {'A': 3, 'B': 5}
    x, y, p = odesys.to_arrays(-37, conc, {'temperature': 298.15})
    fout = odesys.f_cb(x, y, p)
    ref = 3*1e10*np.exp(-40e3/8.3145/298.15)
    assert np.all(abs((fout[:, 0] + ref)/ref) < 1e-14)
    assert np.all(abs((fout[:, 1] - ref)/ref) < 1e-14)


@requires('pyodesys')
def test_get_ode__ArrheniusParam():
    rxn = Reaction({'A': 1}, {'B': 1}, None)
    rxn.param = ArrheniusParam(1e10, 40e3)
    rsys = ReactionSystem([rxn], 'A B')
    odesys = get_odesys(rsys, include_params=True)[0]
    conc = {'A': 3, 'B': 5}
    x, y, p = odesys.to_arrays(-37, conc, {'temperature': 200})
    fout = odesys.f_cb(x, y, p)
    ref = 3*1e10*np.exp(-40e3/8.314472/200)
    assert np.all(abs((fout[:, 0] + ref)/ref) < 1e-14)
    assert np.all(abs((fout[:, 1] - ref)/ref) < 1e-14)


@requires('pyodesys')
def test_get_ode__Radiolytic():
    rad = Radiolytic([2.4e-7])
    rxn = Reaction({'A': 4, 'B': 1}, {'C': 3, 'D': 2}, rad)
    rsys = ReactionSystem([rxn], 'A B C D')
    odesys = get_odesys(rsys, include_params=True)[0]
    c = {'A': 3, 'B': 5, 'C': 11, 'D': 13}
    x, y, p = odesys.to_arrays(-37, c, {'doserate': 0.4, 'density': 0.998})
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7*0.4*0.998
    ref = [-4*r, -r, 3*r, 2*r]
    assert np.all(abs((fout - ref)/ref) < 1e-14)


@requires('pyodesys', units_library)
def test_get_ode__Radiolytic__units():
    rad = Radiolytic([2.4e-7*u.mol/u.joule])
    rxn = Reaction({'A': 4, 'B': 1}, {'C': 3, 'D': 2}, rad)
    rsys = ReactionSystem([rxn], 'A B C D')
    odesys = get_odesys(rsys, include_params=True,
                        unit_registry=SI_base_registry)[0]
    conc = {'A': 3*u.molar, 'B': 5*u.molar, 'C': 11*u.molar, 'D': 13*u.molar}
    x, y, p = odesys.to_arrays(-37*u.second, conc, {
        'doserate': 0.4*u.gray/u.second,
        'density': 0.998*u.kg/u.decimetre**3
    })
    fout = odesys.f_cb(x, y, p)  # f_cb does not carry any units
    r = 2.4e-7*0.4*0.998*1e3  # mol/m3
    ref = [-4*r, -r, 3*r, 2*r]
    assert np.all(abs((fout - ref)/ref) < 1e-14)


@requires('pyodesys', units_library)
def test_get_ode__Radiolytic__units__multi():
    rad = Radiolytic([2.4e-7*u.mol/u.joule])
    rxn = Reaction({'A': 4, 'B': 1}, {'C': 3, 'D': 2}, rad)
    rsys = ReactionSystem([rxn], 'A B C D')
    odesys = get_odesys(rsys, include_params=True,
                        unit_registry=SI_base_registry)[0]
    conc = {'A': 3*u.molar, 'B': 5*u.molar, 'C': 11*u.molar, 'D': 13*u.molar}
    doserates = [dr*u.gray/u.second for dr in [.1, .2, .3, .4]]
    results = odesys.integrate(37*u.second, conc, {
        'doserate': doserates,
        'density': 0.998*u.kg/u.decimetre**3
    })
    assert len(results) == 4
    for i, r in enumerate(results):
        dr = r.params[odesys.param_names.index('doserate')]
        assert dr.ndim == 0 or len(dr) == 1
        assert dr == doserates[i]


class Density(Expr):
    """ Arguments: rho0 drhodT T0 """
    parameter_keys = ('temperature',)
    kw = {}

    def __call__(self, variables, backend=None):
        rho0, drhodT, T0 = self.all_args(variables)
        return rho0 + drhodT*(variables['temperature'] - T0)


@requires('pyodesys')
def test_get_ode__Radiolytic__substitutions():
    rad = Radiolytic([2.4e-7])
    rxn = Reaction({'A': 4, 'B': 1}, {'C': 3, 'D': 2}, rad)
    rsys = ReactionSystem([rxn], 'A B C D')
    substance_rho = Density([1, -1e-3, 273.15])
    odesys = get_odesys(rsys, include_params=True,
                        substitutions={'density': substance_rho})[0]
    conc = {'A': 3, 'B': 5, 'C': 11, 'D': 13}
    state = {'doserate': 0.4, 'temperature': 298.15}
    x, y, p = odesys.to_arrays(-37, conc, state)
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7*0.4*substance_rho({'temperature': 298.15})
    ref = [-4*r, -r, 3*r, 2*r]
    assert np.all(abs((fout - ref)/ref) < 1e-14)


@requires('pyodesys', units_library)
def test_get_ode__Radiolytic__substitutions__units():
    rad = Radiolytic([2.4e-7*u.mol/u.joule])
    rxn = Reaction({'A': 4, 'B': 1}, {'C': 3, 'D': 2}, rad)
    rsys = ReactionSystem([rxn], 'A B C D')
    g_dm3 = u.gram / u.decimetre**3
    kg_dm3 = u.kg / u.decimetre**3
    substance_rho = Density([1*kg_dm3, -1*g_dm3/u.kelvin, 273.15*u.kelvin])
    odesys = get_odesys(rsys, include_params=True, unit_registry=SI_base_registry,
                        substitutions={'density': substance_rho})[0]
    conc = {'A': 3*u.molar, 'B': 5*u.molar, 'C': 11*u.molar, 'D': 13*u.molar}
    x, y, p = odesys.to_arrays(-37*u.second, conc, {'doserate': 0.4*u.gray/u.second, 'temperature': 298.15*u.kelvin})
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7*0.4*0.975 * 1e3  # mol/m3/s
    ref = [-4*r, -r, 3*r, 2*r]
    assert np.all(abs((fout - ref)/ref) < 1e-14)


@requires('pyodesys', units_library)
def test_get_ode__TPoly():
    rate = MassAction(ShiftedTPoly([273.15*u.K, 10/u.molar/u.s, 2/u.molar/u.s/u.K]))
    rxn = Reaction({'A': 1, 'B': 1}, {'C': 3, 'D': 2}, rate, {'A': 3})
    rsys = ReactionSystem([rxn], 'A B C D')
    odesys = get_odesys(rsys, unit_registry=SI_base_registry)[0]
    conc = {'A': 3*u.molar, 'B': 5*u.molar, 'C': 11*u.molar, 'D': 13*u.molar}
    x, y, p = odesys.to_arrays(-37*u.second, conc, {'temperature': 298.15*u.kelvin})
    fout = odesys.f_cb(x, y, p)
    r = 3*5*(10+2*25)*1000  # mol/m3/s
    ref = [-4*r, -r, 3*r, 2*r]
    assert np.all(abs((fout - ref)/ref) < 1e-14)


@requires('pyodesys', units_library)
def test_get_odesys__time_dep_rate():

    class RampedRate(Expr):
        argument_names = ('rate_constant', 'ramping_rate')

        def __call__(self, variables, reaction, backend=math):
            rate_constant, ramping_rate = self.all_args(variables, backend=backend)
            return rate_constant * ramping_rate * variables['time']

    rate = MassAction(RampedRate([7, 2]))
    rxn = Reaction({'A': 1}, {'B': 3}, rate)
    rsys = ReactionSystem([rxn], 'A B')
    odesys = get_odesys(rsys)[0]
    conc = {'A': 3, 'B': 11}
    x, y, p = odesys.to_arrays([5, 13, 17], conc, ())
    fout = odesys.f_cb(x, y, p)
    r = 2*7*3
    ref = np.array([
        [-r*5, -r*13, -r*17],
        [r*5*3, r*13*3, r*17*3]
    ]).T
    assert np.allclose(fout, ref)


@pytest.mark.slow
@requires('pyodesys', units_library)
def test_get_odesys__time_dep_temperature():
    import sympy as sp

    def refA(t, A0, A, Ea_over_R, T0, dTdt):
        T = (T0 + dTdt*t)
        d_Ei = sp.Ei(-Ea_over_R/T0).n(100).round(90) - sp.Ei(-Ea_over_R/T).n(100).round(90)
        d_Texp = T0*sp.exp(-Ea_over_R/T0) - T*sp.exp(-Ea_over_R/T)
        return A0*sp.exp(A/dTdt*(Ea_over_R*d_Ei + d_Texp)).n(30)

    params = A0, A, Ea_over_R, T0, dTdt = [13, 1e10, 56e3/8, 273, 2]
    B0 = 11
    rate = MassAction(Arrhenius([A, Ea_over_R]))
    rxn = Reaction({'A': 1}, {'B': 3}, rate)
    rsys = ReactionSystem([rxn], 'A B')
    rt = RampedTemp([T0, dTdt], ('init_temp', 'ramp_rate'))
    odesys, extra = get_odesys(rsys, False, substitutions={'temperature': rt})
    all_pk, unique, p_units = map(extra.get, 'param_keys unique p_units'.split())
    conc = {'A': A0, 'B': B0}
    tout = [2, 5, 10]

    for ramp_rate in [2, 3, 4]:
        unique['ramp_rate'] = ramp_rate
        xout, yout, info = odesys.integrate(10, conc, unique, atol=1e-10, rtol=1e-12)
        params[-1] = ramp_rate
        Aref = np.array([float(refA(t, *params)) for t in xout])
        # Aref = 1/(1/13 + 2*1e-6*t_unitless)
        yref = np.zeros((xout.size, 2))
        yref[:, 0] = Aref
        yref[:, 1] = B0 + 3*(A0-Aref)
        assert allclose(yout, yref)

    unique['ramp_rate'] = 2
    x, y, p = odesys.to_arrays(tout, conc, unique)
    fout = odesys.f_cb(x, y, p)

    def r(t):
        return A*np.exp(-Ea_over_R/(T0 + dTdt*t))*A0  # refA(t, *params)

    ref = np.array([
        [-r(2), -r(5), -r(10)],
        [3*r(2), 3*r(5), 3*r(10)]
    ]).T
    assert np.allclose(fout, ref)


@requires('numpy', 'pyodesys')
def test_get_odesys__late_binding():
    def _gibbs(args, T, R, backend, **kwargs):
        H, S = args
        return backend.exp(-(H - T*S)/(R*T))

    def _eyring(args, T, R, k_B, h, backend, **kwargs):
        H, S = args
        return k_B/h*T*backend.exp(-(H - T*S)/(R*T))

    gibbs_pk = ('temperature', 'molar_gas_constant')
    eyring_pk = gibbs_pk + ('Boltzmann_constant', 'Planck_constant')

    GibbsEC = MassActionEq.from_callback(_gibbs, argument_names=('H', 'S'), parameter_keys=gibbs_pk)
    EyringMA = MassAction.from_callback(_eyring, argument_names=('H', 'S'), parameter_keys=eyring_pk)

    uk_equil = ('He_assoc', 'Se_assoc')
    beta = GibbsEC(unique_keys=uk_equil)  # equilibrium parameters

    uk_kinet = ('Ha_assoc', 'Sa_assoc')
    bimol_barrier = EyringMA(unique_keys=uk_kinet)  # activation parameters

    eq = Equilibrium({'Fe+3', 'SCN-'}, {'FeSCN+2'}, beta)
    rsys = ReactionSystem(eq.as_reactions(kf=bimol_barrier))
    odesys, extra = get_odesys(rsys, include_params=False)
    pk, unique, p_units = map(extra.get, 'param_keys unique p_units'.split())
    assert sorted(unique) == sorted(uk_equil + uk_kinet)
    assert sorted(pk) == sorted(eyring_pk)


@requires('numpy', 'pyodesys')
def test_get_odesys__ScaledSys():
    from pyodesys.symbolic import ScaledSys
    k = .2
    a = Substance('A')
    b = Substance('B')
    r = Reaction({'A': 1}, {'B': 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ['A', 'B']
    odesys = get_odesys(rsys, include_params=True, SymbolicSys=ScaledSys)[0]
    c0 = {
        'A': 1.0,
        'B': 3.0,
    }
    t = np.linspace(0.0, 10.0)
    xout, yout, info = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k*t)
    yref[:, 1] = 4 - np.exp(-k*t)
    assert np.allclose(yout, yref)


@requires('numpy', 'pyodesys', 'sympy')
def test_get_odesys__max_euler_step_cb():
    rsys = ReactionSystem.from_string('\n'.join(['H2O -> H+ + OH-; 1e-4', 'OH- + H+ -> H2O; 1e10']))
    odesys, extra = get_odesys(rsys)
    r1 = 1.01e-4
    r2 = 6e-4
    dH2Odt = r2 - r1
    euler_ref = 2e-7/dH2Odt
    assert abs(extra['max_euler_step_cb'](0, {'H2O': 1.01, 'H+': 2e-7, 'OH-': 3e-7}) - euler_ref)/euler_ref < 1e-8


@requires('numpy', 'pyodesys', 'sympy')
@pytest.mark.parametrize('substances', permutations(['H2O', 'H+', 'OH-']))
def test_get_odesys__linear_dependencies__preferred(substances):
    rsys = ReactionSystem.from_string('\n'.join(['H2O -> H+ + OH-; 1e-4', 'OH- + H+ -> H2O; 1e10']), substances)
    assert isinstance(rsys.substances, OrderedDict)
    odesys, extra = get_odesys(rsys)

    af_H2O_H = extra['linear_dependencies'](['H+', 'H2O'])
    import sympy
    y0 = {k: sympy.Symbol(k+'0') for k in rsys.substances}
    af_H2O_H(None, {odesys[k]: v for k, v in y0.items()}, None, sympy)  # ensure idempotent
    exprs_H2O_H = af_H2O_H(None, {odesys[k]: v for k, v in y0.items()}, None, sympy)
    ref_H2O_H = {
        'H2O': y0['H2O'] + y0['OH-'] - odesys['OH-'],  # oxygen
        'H+': 2*y0['H2O'] + y0['H+'] + y0['OH-'] - odesys['OH-'] - 2*(
            y0['H2O'] + y0['OH-'] - odesys['OH-'])  # hydrogen
    }
    for k, v in ref_H2O_H.items():
        assert (exprs_H2O_H[odesys[k]] - v) == 0


@requires('numpy', 'pyodesys', 'sympy', 'pycvodes')
@pytest.mark.parametrize('preferred', [None, ['H+', 'OH-'], ['H2O', 'H+'], ['H2O', 'OH-']])
def test_get_odesys__linear_dependencies__PartiallySolvedSystem(preferred):
    import sympy
    from pyodesys.symbolic import PartiallySolvedSystem
    rsys = ReactionSystem.from_string('\n'.join(['H2O -> H+ + OH-; 1e-4', 'OH- + H+ -> H2O; 1e10']))
    odesys, extra = get_odesys(rsys)
    c0 = {'H2O': 0, 'H+': 2e-7, 'OH-': 3e-7}
    h0max = extra['max_euler_step_cb'](0, c0)
    analytic_factory = extra['linear_dependencies']()
    y0 = {k: sympy.Symbol(k+'0') for k in rsys.substances}
    analytic_factory(None, {odesys[k]: v for k, v in y0.items()}, None, sympy)
    psys = PartiallySolvedSystem(odesys, analytic_factory)
    xout, yout, info = psys.integrate(1, c0, atol=1e-12, rtol=1e-10, first_step=h0max*1e-12,
                                      integrator='cvode')
    c_reac = c0['H+'], c0['OH-']
    H2O_ref = binary_rev(xout, 1e10, 1e-4, c0['H2O'], max(c_reac), min(c_reac))
    assert np.allclose(yout[:, psys.names.index('H2O')], H2O_ref)
    assert np.allclose(yout[:, psys.names.index('H+')], c0['H+'] + c0['H2O'] - H2O_ref)
    assert np.allclose(yout[:, psys.names.index('OH-')], c0['OH-'] + c0['H2O'] - H2O_ref)


@requires('numpy', 'pyodesys', 'sympy', 'pycvodes')
def test_get_odesys__Equilibrium_as_reactions():
    from chempy import Equilibrium, ReactionSystem
    eq = Equilibrium({'Fe+3', 'SCN-'}, {'FeSCN+2'}, 10**2)
    substances = 'Fe+3 SCN- FeSCN+2'.split()
    rsys = ReactionSystem(eq.as_reactions(kf=3.0), substances)
    odesys, extra = get_odesys(rsys)
    init_conc = {'Fe+3': 1.0, 'SCN-': .3, 'FeSCN+2': 0}
    tout, Cout, info = odesys.integrate(5, init_conc, integrator='cvode', atol=1e-11, rtol=1e-12)
    cmplx_ref = binary_rev(tout, 3, 3.0/100, init_conc['FeSCN+2'], init_conc['Fe+3'], init_conc['SCN-'])
    assert np.allclose(Cout[:, 2], cmplx_ref)


@requires('numpy', 'pyodesys', 'sympy', 'pycvodes')
def test_get_odesys__Expr_as_param():
    def _eyring_pe(args, T, backend=math, **kwargs):
        freq, = args
        return freq*T

    EyringPreExp = Expr.from_callback(_eyring_pe, argument_names=('freq',),
                                      parameter_keys=('temperature',))

    def _k(args, T, backend=math, **kwargs):
        A, H, S = args
        return A*backend.exp(-(H - T*S)/(8.314511*T))

    EyringMA = MassAction.from_callback(_k, parameter_keys=('temperature',),
                                        argument_names=('Aa', 'Ha', 'Sa'))
    kb_h = 2.08e10
    rxn = Reaction({'A'}, {'B'}, EyringMA(unique_keys=('A_u', 'H_u', 'S_u')))
    rsys = ReactionSystem([rxn], ['A', 'B'])
    odesys, extra = get_odesys(rsys, include_params=False, substitutions={'A_u': EyringPreExp(kb_h)})
    y0 = defaultdict(float, {'A': 7.0})
    rt = 293.15
    xout, yout, info = odesys.integrate(5, y0, {'H_u': 117e3, 'S_u': 150, 'temperature': rt},
                                        integrator='cvode', atol=1e-12, rtol=1e-10, nsteps=1000)
    kref = kb_h*rt*np.exp(-(117e3 - rt*150)/(8.314511*rt))
    ref = y0['A']*np.exp(-kref*xout)
    assert np.allclose(yout[:, 0], ref)
    assert np.allclose(yout[:, 1], y0['A'] - ref)


@requires('numpy', 'pyodesys', 'sympy', 'pycvodes')
def test_get_odesys__Expr_as_param__unique_as_param():
    def _eyring_pe_coupled(args, T, S, backend=math, **kwargs):
        freq, = args
        return freq*T/S

    EyringPreExpCoupled = Expr.from_callback(_eyring_pe_coupled, argument_names=('freq',),
                                             parameter_keys=('temperature', 'S_u'))

    def _k(args, T, backend=math, **kwargs):
        A, H, S = args
        return A*backend.exp(-(H - T*S)/(8.314511*T))

    EyringMA = MassAction.from_callback(_k, parameter_keys=('temperature',),
                                        argument_names=('Aa', 'Ha', 'Sa'))
    kb_h = 2.08e10
    rxn = Reaction({'A'}, {'B'}, EyringMA(unique_keys=('A_u', 'H_u', 'S_u')))
    rsys = ReactionSystem([rxn], ['A', 'B'])
    odesys2, extra2 = get_odesys(rsys, include_params=False,
                                 substitutions={'A_u': EyringPreExpCoupled(kb_h)})
    y0 = defaultdict(float, {'A': 7.0})
    rt = 293.15
    xout2, yout2, info2 = odesys2.integrate(5, y0, {'H_u': 107e3, 'S_u': 150, 'temperature': rt},
                                            integrator='cvode', atol=1e-12, rtol=1e-10, nsteps=1000)
    kref2 = kb_h*rt*np.exp(-(107e3 - rt*150)/(8.314511*rt))/150
    ref2 = y0['A']*np.exp(-kref2*xout2)
    assert np.allclose(yout2[:, 0], ref2)
    assert np.allclose(yout2[:, 1], y0['A'] - ref2)


@requires('pyodesys', 'pycvodes')
def test_chained_parameter_variation():
    ratex = MassAction(Arrhenius([1e10, 63e3/8.3145]))
    rxn = Reaction({'A': 1}, {'B': 1}, ratex)
    rsys = ReactionSystem([rxn], 'A B')
    odesys, extra = get_odesys(rsys, include_params=False)
    param_keys, unique_keys, p_units = map(extra.get, 'param_keys unique p_units'.split())
    conc = {'A': 3.17, 'B': 5.03}
    Ts = (294, 304, 317)
    times = [3.1, 2.1, 5.3]
    kw = dict(integrator='cvode', atol=1e-12, rtol=1e-13, first_step=1e-14)
    tout, cout, info = chained_parameter_variation(
        odesys, times, conc, {'temperature': Ts}, {}, integrate_kwargs=kw)
    assert len(info['nfev']) == 3
    assert info['nfev'][0] > 2
    assert info['nfev'][1] > 2
    assert info['nfev'][2] > 2
    assert np.all(np.diff(tout) > 0)
    tout1 = tout[tout <= times[0]]
    tout23 = tout[tout > times[0]]
    tout2 = tout23[tout23 <= times[0] + times[1]]
    tout3 = tout23[tout23 > times[0] + times[1]]

    def _ref(y0, x, T, x0):
        k = 1e10*np.exp(-63e3/8.3145/T)
        return y0*np.exp(-k*(x-x0))

    Aref1 = _ref(conc['A'], tout1, Ts[0], tout1[0])
    Bref1 = conc['B'] + conc['A'] - Aref1

    Aref2 = _ref(Aref1[-1], tout2, Ts[1], tout1[-1])
    Bref2 = Bref1[-1] + Aref1[-1] - Aref2

    Aref3 = _ref(Aref2[-1], tout3, Ts[2], tout2[-1])
    Bref3 = Bref2[-1] + Aref2[-1] - Aref3

    cref = np.concatenate([np.vstack((a, b)).T for a, b in [(Aref1, Bref1), (Aref2, Bref2), (Aref3, Bref3)]])
    forgive = 27*1.1
    assert np.allclose(cref, cout, atol=kw['atol']*forgive, rtol=kw['rtol']*forgive)


@requires('pyodesys', 'scipy', 'sym')
def test_get_odesys__cstr():
    rsys = ReactionSystem.from_string("2 H2O2 -> O2 + 2 H2O; 5")
    odesys, extra = get_odesys(rsys, cstr=True)
    fr, fc = extra['cstr_fr_fc']
    tout, c0 = np.linspace(0, .13, 7), {'H2O2': 2, 'O2': 4, 'H2O': 3}
    params = {fr: 13, fc['H2O2']: 11, fc['O2']: 43, fc['H2O']: 45}
    res = odesys.integrate(tout, c0, params)
    from chempy.kinetics.integrated import binary_irrev_cstr

    def get_analytic(result, k, n):
        ref = binary_irrev_cstr(result.xout, 5, result.named_dep('H2O2')[0],
                                result.named_dep(k)[0], result.named_param(fc['H2O2']), result.named_param(fc[k]),
                                result.named_param(fr), n)
        return np.array(ref).T

    ref_O2 = get_analytic(res, 'O2', 1)
    ref_H2O = get_analytic(res, 'H2O', 2)
    assert np.allclose(res.named_dep('H2O2'), ref_O2[:, 0])
    assert np.allclose(res.named_dep('H2O2'), ref_H2O[:, 0])
    assert np.allclose(res.named_dep('O2'), ref_O2[:, 1])
    assert np.allclose(res.named_dep('H2O'), ref_H2O[:, 1])


@requires('pygslodeiv2', 'sym', units_library)
def test_get_odesys_rsys_with_units():
    rsys = ReactionSystem.from_string("""
    A -> B; 0.096/s
    B + C -> P; 4e3/M/s
    """, substance_factory=Substance)
    with pytest.raises(Exception):
        get_odesys(rsys)  # not a strict test, SI_base_registry could be made the default

    odesys, extra = get_odesys(rsys, unit_registry=SI_base_registry)
    tend = 10
    tend_units = tend*u.s
    c0 = {'A': 1e-6, 'B': 0, 'C': 1, 'P': 0}
    c0_units = {k: v*u.molar for k, v in c0.items()}
    result1 = odesys.integrate(tend_units, c0_units, integrator='gsl')
    assert result1.info['success']

    with pytest.raises(Exception):
        odesys.integrate(tend, c0, integrator='gsl')


@requires('pyodeint', 'sym', units_library)
def test_get_odesys_rsys_with_units__named_params():
    rsys = ReactionSystem.from_string("""
    A -> B; 'k1'
    B + C -> P; 'k2'
    """, substance_factory=Substance)
    odesys, extra = get_odesys(rsys, include_params=False, unit_registry=SI_base_registry)
    tend = 10
    tend_units = tend*u.s
    c0 = {'A': 1e-6, 'B': 0, 'C': 1, 'P': 0}
    p = {'k1': 3, 'k2': 4}
    p_units = {'k1': 3/u.s, 'k2': 4/u.M/u.s}
    c0_units = {k: v*u.molar for k, v in c0.items()}
    result1 = odesys.integrate(tend_units, c0_units, p_units, integrator='odeint')
    assert result1.info['success']

    with pytest.raises(Exception):
        odesys.integrate(tend, c0, p, integrator='odeint')


@requires('pycvodes', 'sym', units_library)
def test_get_odesys__Eyring():
    R = 8.314472
    T_K = 300
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string("""
    NOBr -> NO + Br; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(dH=dH, dS=dS), substances='NOBr NO Br'.split())
    kref = 20836643994.118652*T_K*np.exp(-(dH - T_K*dS)/(R*T_K))
    NOBr0_M = 0.7
    init_cond = dict(
        NOBr=NOBr0_M*u.M,
        NO=0*u.M,
        Br=0*u.M
    )
    t = 5*u.second
    params = dict(
        temperature=T_K*u.K
    )

    def check(rsys):
        odesys, extra = get_odesys(rsys, unit_registry=SI_base_registry, constants=const)
        res = odesys.integrate(t, init_cond, params, integrator='cvode')
        NOBr_ref = NOBr0_M*np.exp(-kref*to_unitless(res.xout, u.second))
        ref = np.array([NOBr_ref] + [NOBr0_M - NOBr_ref]*2).T
        cmp = to_unitless(res.yout, u.M)
        assert np.allclose(cmp, ref)
    check(rsys1)
    rsys2 = ReactionSystem.from_string("""
    NOBr -> NO + Br; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(dH=dH, dS=dS), substances='NOBr NO Br'.split())
    check(rsys2)


@requires('pycvodes', 'sym', units_library)
def test_get_odesys__Eyring_2nd_order():
    R = 8.314472
    T_K = 300
    dH = 80e3
    dS = 10
    rsys1b = ReactionSystem.from_string("""
    NO + Br -> NOBr; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(dH=dH, dS=dS))
    c0 = 1  # mol/dm3 === 1000 mol/m3
    kbref = 20836643994.118652*T_K*np.exp(-(dH - T_K*dS)/(R*T_K))/c0
    NO0_M = 1.5
    Br0_M = 0.7
    init_cond = dict(
        NOBr=0*u.M,
        NO=NO0_M*u.M,
        Br=Br0_M*u.M
    )
    t = 5*u.second
    params = dict(
        temperature=T_K*u.K
    )

    def analytic_b(t):
        U, V = NO0_M, Br0_M
        d = U - V
        return (U*(1 - np.exp(-kbref*t*d)))/(U/V - np.exp(-kbref*t*d))

    def check(rsys):
        odesys, extra = get_odesys(rsys, unit_registry=SI_base_registry, constants=const)
        res = odesys.integrate(t, init_cond, params, integrator='cvode')
        t_sec = to_unitless(res.xout, u.second)
        NOBr_ref = analytic_b(t_sec)
        cmp = to_unitless(res.yout, u.M)
        ref = np.empty_like(cmp)
        ref[:, odesys.names.index('NOBr')] = NOBr_ref
        ref[:, odesys.names.index('Br')] = Br0_M - NOBr_ref
        ref[:, odesys.names.index('NO')] = NO0_M - NOBr_ref
        assert np.allclose(cmp, ref)

    check(rsys1b)
    rsys2b = ReactionSystem.from_string("""
    NO + Br -> NOBr; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(dH=dH, dS=dS))
    check(rsys2b)


@requires('pycvodes', 'sym', 'scipy', units_library)
def test_get_odesys__Eyring_1st_order_linearly_ramped_temperature():
    from scipy.special import expi

    def analytic_unit0(t, T0, dH, dS):
        R = 8.314472
        kB = 1.3806504e-23
        h = 6.62606896e-34
        A = kB/h*np.exp(dS/R)
        B = dH/R
        return np.exp(A*(
            (-B**2*np.exp(B/T0)*expi(-B/T0) - T0*(B - T0))*np.exp(-B/T0) +
            (B**2*np.exp(B/(t + T0))*expi(-B/(t + T0)) - (t + T0)*(-B + t + T0))*np.exp(-B/(t + T0))
        )/2)

    T_K = 290
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string("""
    NOBr -> NO + Br; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(dH=dH, dS=dS))

    NOBr0_M = 0.7
    init_cond = dict(
        NOBr=NOBr0_M*u.M,
        NO=0*u.M,
        Br=0*u.M
    )
    t = 20*u.second

    def check(rsys):
        odes, extra = get_odesys(rsys, unit_registry=SI_base_registry, constants=const, substitutions={
            'temperature': RampedTemp([T_K*u.K, 1*u.K/u.s])})
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, integrator='cvode')
            t_sec = to_unitless(res.xout, u.second)
            NOBr_ref = NOBr0_M*analytic_unit0(t_sec, T_K, dH, dS)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index('NOBr')] = NOBr_ref
            ref[:, odesys.names.index('Br')] = NOBr0_M - NOBr_ref
            ref[:, odesys.names.index('NO')] = NOBr0_M - NOBr_ref
            assert np.allclose(cmp, ref)

    check(rsys1)
    rsys2 = ReactionSystem.from_string("""
    NOBr -> NO + Br; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(dH=dH, dS=dS))
    check(rsys2)


@requires('pycvodes', 'sym', 'scipy', units_library)
def test_get_odesys__Eyring_2nd_order_linearly_ramped_temperature():
    from scipy.special import expi

    def analytic_unit0(t, k, m, dH, dS):
        R = 8.314472
        kB = 1.3806504e-23
        h = 6.62606896e-34
        A = kB/h*np.exp(dS/R)
        B = dH/R
        return k*np.exp(B*(k*t + 2*m)/(m*(k*t + m)))/(
            A*(-B**2*np.exp(B/(k*t + m))*expi(-B/(k*t + m)) - B*k*t - B*m + k**2*t**2 + 2*k*m*t + m**2)*np.exp(B/m) +
            (A*B**2*np.exp(B/m)*expi(-B/m) - A*m*(-B + m) + k*np.exp(B/m))*np.exp(B/(k*t + m))
        )

    T_K = 290
    dTdt_Ks = 3
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string("""
    2 NO2 -> N2O4; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(dH=dH, dS=dS))

    NO2_M = 1.0
    init_cond = dict(
        NO2=NO2_M*u.M,
        N2O4=0*u.M
    )
    t = 20*u.second

    def check(rsys):
        odes, extra = get_odesys(rsys, unit_registry=SI_base_registry, constants=const, substitutions={
            'temperature': RampedTemp([T_K*u.K, dTdt_Ks*u.K/u.s])})
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, integrator='cvode')
            t_sec = to_unitless(res.xout, u.second)
            NO2_ref = analytic_unit0(t_sec, dTdt_Ks, T_K, dH, dS)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index('NO2')] = NO2_ref
            ref[:, odesys.names.index('N2O4')] = (NO2_M - NO2_ref)/2
            assert np.allclose(cmp, ref)

    check(rsys1)
    rsys2 = ReactionSystem.from_string("""
    2 NO2 -> N2O4; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(dH=dH, dS=dS))
    check(rsys2)


@requires('pycvodes', 'sym', units_library)
def test_get_odesys__Eyring_2nd_order_reversible():
    R = 8.314472
    T_K = 273.15 + 20  # 20 degree celsius
    kB = 1.3806504e-23
    h = 6.62606896e-34

    dHf = 74e3
    dSf = R*np.log(h/kB/T_K*1e16)

    dHb = 79e3
    dSb = dSf - 23

    rsys1 = ReactionSystem.from_string("""
    Fe+3 + SCN- -> FeSCN+2; EyringParam(dH={dHf}*J/mol, dS={dSf}*J/K/mol)
    FeSCN+2 -> Fe+3 + SCN-; EyringParam(dH={dHb}*J/mol, dS={dSb}*J/K/mol)
    """.format(dHf=dHf, dSf=dSf, dHb=dHb, dSb=dSb))
    kf_ref = 20836643994.118652*T_K*np.exp(-(dHf - T_K*dSf)/(R*T_K))
    kb_ref = 20836643994.118652*T_K*np.exp(-(dHb - T_K*dSb)/(R*T_K))
    Fe0 = 6e-3
    SCN0 = 2e-3
    init_cond = {
        'Fe+3': Fe0*u.M,
        'SCN-': SCN0*u.M,
        'FeSCN+2': 0*u.M
    }
    t = 3*u.second

    def check(rsys, params):
        odes, extra = get_odesys(rsys, include_params=False,
                                 unit_registry=SI_base_registry, constants=const)
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, params, integrator='cvode')
            t_sec = to_unitless(res.xout, u.second)
            FeSCN_ref = binary_rev(t_sec, kf_ref, kb_ref, 0, Fe0, SCN0)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index('FeSCN+2')] = FeSCN_ref
            ref[:, odesys.names.index('Fe+3')] = Fe0 - FeSCN_ref
            ref[:, odesys.names.index('SCN-')] = SCN0 - FeSCN_ref
            assert np.allclose(cmp, ref)

    check(rsys1, {'temperature': T_K*u.K})
    rsys2 = ReactionSystem.from_string("""
    Fe+3 + SCN- -> FeSCN+2; MassAction(EyringHS([{dHf}*J/mol, {dSf}*J/K/mol]))
    FeSCN+2 -> Fe+3 + SCN-; MassAction(EyringHS([{dHb}*J/mol, {dSb}*J/K/mol]))
    """.format(dHf=dHf, dSf=dSf, dHb=dHb, dSb=dSb))
    check(rsys2, {'temperature': T_K*u.K})
    rsys3 = ReactionSystem.from_string("""
    Fe+3 + SCN- -> FeSCN+2; MassAction(EyringHS.fk('dHf', 'dSf'))
    FeSCN+2 -> Fe+3 + SCN-; MassAction(EyringHS.fk('dHb', 'dSb'))
    """)
    check(rsys3, dict(temperature=T_K*u.K,
                      dHf=dHf*u.J/u.mol, dSf=dSf*u.J/u.mol/u.K,
                      dHb=dHb*u.J/u.mol, dSb=dSb*u.J/u.mol/u.K))


@requires('numpy', 'pyodesys', 'sympy', 'pycvodes')
def test_create_odesys():
    rsys = ReactionSystem.from_string("""
    A -> B; 'k1'
    B + C -> P; 'k2'
    """, substance_factory=Substance)

    odesys, odesys_extra = create_odesys(rsys, unit_registry=SI_base_registry)

    tend = 10
    c0 = {'A': 1e-6, 'B': 0, 'C': 1, 'P': 0}
    p = dict(k1=3, k2=4)

    tend_units = tend*u.s
    p_units = {'k1': p['k1']/u.s, 'k2': p['k2']/u.M/u.s}
    c0_units = {k: v*u.molar for k, v in c0.items()}

    validation = odesys_extra['validate'](dict(c0_units, **p_units))
    P, = validation['not_seen']
    assert P.name == 'P'

    ref_rates = {'A': -p_units['k1']*c0_units['A'], 'P': p_units['k2']*c0_units['B']*c0_units['C']}
    ref_rates['B'] = -ref_rates['A'] - ref_rates['P']
    ref_rates['C'] = -ref_rates['P']
    assert validation['rates'] == ref_rates

    result1, result1_extra = odesys_extra['unit_aware_solve'](
        tend_units, c0_units, p_units, integrator='cvode')
    assert result1.info['success']

    result2 = odesys.integrate(tend, c0, p, integrator='cvode')
    assert np.allclose(result2.yout[-1, :], to_unitless(result1.yout[-1, :], u.molar))


@requires('pycvodes', 'sym', units_library)
def test_create_odesys__Radiolytic():
    rsys1 = ReactionSystem.from_string("""
    -> e-(aq); Radiolytic.fk('g_emaq')
    """, checks=())
    ic1 = {'e-(aq)': 0.0}
    t1 = 5
    p1 = dict(
        g_emaq=42.0,
        doserate=17.0,
        density=5.0
    )
    odesys1, odesys_extra = create_odesys(rsys1)
    result1 = odesys1.integrate(t1, ic1, p1)
    yref1 = result1.xout*p1['g_emaq']*p1['doserate']*p1['density']
    assert np.allclose(yref1, result1.yout.squeeze())



# -*- coding: utf-8 -*-
"""
Module collecting classes and functions for dealing with (multiphase) chemical
equilibria.

.. Note::

  This module is provisional at the moment, i.e. the API is not stable and may
  break without a deprecation cycle.

"""
from __future__ import division, absolute_import

import warnings

import numpy as np

from .chemistry import equilibrium_quotient, Equilibrium, Species
from .reactionsystem import ReactionSystem
from ._util import get_backend
from .util.pyutil import deprecated
from ._eqsys import EqCalcResult, NumSysLin, NumSysLog, NumSysSquare as _NumSysSquare


NumSysSquare = deprecated()(_NumSysSquare)


class EqSystem(ReactionSystem):

    _BaseReaction = Equilibrium
    _BaseSubstance = Species

    def html(self, *args, **kwargs):
        k = 'color_categories'
        kwargs[k] = kwargs.get(k, False)
        return super(EqSystem, self).html(*args, **kwargs)

    def eq_constants(self, non_precip_rids=(), eq_params=None, small=0):
        if eq_params is None:
            eq_params = [eq.param for eq in self.rxns]
        return [small if idx in non_precip_rids else
                eq for idx, eq in enumerate(eq_params)]

    def equilibrium_quotients(self, concs):
        stoichs = self.stoichs()
        return [equilibrium_quotient(concs, stoichs[ri, :])
                for ri in range(self.nr)]

    def stoichs_constants(self, eq_params=None, rref=False, Matrix=None,
                          backend=None, non_precip_rids=()):
        if eq_params is None:
            eq_params = self.eq_constants()
        if rref:
            from pyneqsys.symbolic import linear_rref
            be = get_backend(backend)
            rA, rb = linear_rref(self.stoichs(non_precip_rids),
                                 list(map(be.log, eq_params)),
                                 Matrix)
            return rA.tolist(), list(map(be.exp, rb))
        else:
            return (self.stoichs(non_precip_rids),
                    eq_params)

    def composition_conservation(self, concs, init_concs):
        composition_vecs, comp_keys = self.composition_balance_vectors()
        A = np.array(composition_vecs)
        return (comp_keys,
                np.dot(A, self.as_per_substance_array(concs).T),
                np.dot(A, self.as_per_substance_array(init_concs).T))

    def other_phase_species_idxs(self, phase_idx=0):
        return [idx for idx, s in enumerate(
            self.substances.values()) if s.phase_idx != phase_idx]

    @property
    @deprecated(last_supported_version='0.3.1', will_be_missing_in='0.8.0',
                use_instead=other_phase_species_idxs)
    def precipitate_substance_idxs(self):
        return [idx for idx, s in enumerate(
            self.substances.values()) if s.precipitate]

    def phase_transfer_reaction_idxs(self, phase_idx=0):
        return [idx for idx, rxn in enumerate(self.rxns)
                if rxn.has_precipitates(self.substances)]

    @property
    @deprecated(last_supported_version='0.3.1', will_be_missing_in='0.8.0',
                use_instead=phase_transfer_reaction_idxs)
    def precipitate_rxn_idxs(self):
        return [idx for idx, rxn in enumerate(self.rxns)
                if rxn.has_precipitates(self.substances)]

    def dissolved(self, concs):
        """ Return dissolved concentrations """
        new_concs = concs.copy()
        for r in self.rxns:
            if r.has_precipitates(self.substances):
                net_stoich = np.asarray(r.net_stoich(self.substances))
                s_net, s_stoich, s_idx = r.precipitate_stoich(self.substances)
                new_concs -= new_concs[s_idx]/s_stoich * net_stoich
        return new_concs

    def _fw_cond_factory(self, ri, rtol=1e-14):
        rxn = self.rxns[ri]

        def fw_cond(x, p):
            precip_stoich_coeff, precip_idx = rxn.precipitate_stoich(self.substances)[1:3]
            q = rxn.Q(self.substances, self.dissolved(x))
            k = rxn.equilibrium_constant()
            if precip_stoich_coeff > 0:
                return q*(1+rtol) < k
            elif precip_stoich_coeff < 0:
                return q > k*(1+rtol)
            else:
                raise NotImplementedError
        return fw_cond

    def _bw_cond_factory(self, ri, small):
        rxn = self.rxns[ri]

        def bw_cond(x, p):
            precipitate_idx = rxn.precipitate_stoich(self.substances)[2]
            if x[precipitate_idx] < small:
                return False
            else:
                return True
        return bw_cond

    def _SymbolicSys_from_NumSys(self, NS, conds, rref_equil, rref_preserv,
                                 new_eq_params=True):
        from pyneqsys.symbolic import SymbolicSys
        import sympy as sp
        ns = NS(self, backend=sp, rref_equil=rref_equil,
                rref_preserv=rref_preserv, precipitates=conds,
                new_eq_params=new_eq_params)
        symb_kw = {}
        if ns.pre_processor is not None:
            symb_kw['pre_processors'] = [ns.pre_processor]
        if ns.post_processor is not None:
            symb_kw['post_processors'] = [ns.post_processor]
        if ns.internal_x0_cb is not None:
            symb_kw['internal_x0_cb'] = ns.internal_x0_cb
        return SymbolicSys.from_callback(
            ns.f, self.ns, nparams=self.ns + (self.nr if new_eq_params else 0),
            **symb_kw)

    def get_neqsys_conditional_chained(self, rref_equil=False,
                                       rref_preserv=False, NumSys=NumSysLin, **kwargs):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def factory(conds):
            return ChainedNeqSys([self._SymbolicSys_from_NumSys(
                NS, conds, rref_equil, rref_preserv, **kwargs
            ) for NS in NumSys])

        cond_cbs = [(self._fw_cond_factory(ri),
                     self._bw_cond_factory(ri, NumSys[0].small)) for
                    ri in self.phase_transfer_reaction_idxs()]
        return ConditionalNeqSys(cond_cbs, factory)

    def get_neqsys_chained_conditional(self, rref_equil=False,
                                       rref_preserv=False,
                                       NumSys=NumSysLin, **kwargs):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def mk_factory(NS):
            def factory(conds):
                return self._SymbolicSys_from_NumSys(NS, conds, rref_equil,
                                                     rref_preserv, **kwargs)
            return factory

        return ChainedNeqSys(
            [ConditionalNeqSys(
                [(self._fw_cond_factory(ri),
                  self._bw_cond_factory(ri, NS.small)) for
                 ri in self.phase_transfer_reaction_idxs()],
                mk_factory(NS)
            ) for NS in NumSys])

    def get_neqsys_static_conditions(self, rref_equil=False,
                                     rref_preserv=False,
                                     NumSys=(NumSysLin,), precipitates=None, **kwargs):
        if precipitates is None:
            precipitates = (False,)*len(self.phase_transfer_reaction_idxs())
        from pyneqsys import ChainedNeqSys
        return ChainedNeqSys([self._SymbolicSys_from_NumSys(
            NS, precipitates, rref_equil, rref_preserv, **kwargs) for NS in NumSys])

    def get_neqsys(self, neqsys_type, NumSys=NumSysLin, **kwargs):
        new_kw = {'rref_equil': False, 'rref_preserv': False}
        if neqsys_type == 'static_conditions':
            new_kw['precipitates'] = None
        for k in new_kw:
            if k in kwargs:
                new_kw[k] = kwargs.pop(k)

        try:
            NumSys[0]
        except TypeError:
            new_kw['NumSys'] = (NumSys,)
        else:
            new_kw['NumSys'] = NumSys

        return getattr(self, 'get_neqsys_' + neqsys_type)(**new_kw)

    def non_precip_rids(self, precipitates):
        return [idx for idx, precip in zip(
            self.phase_transfer_reaction_idxs(), precipitates) if not precip]

    def _result_is_sane(self, init_concs, x, rtol=1e-9):
        sc_upper_bounds = np.array(self.upper_conc_bounds(init_concs))
        neg_conc, too_much = np.any(x < 0), np.any(
            x > sc_upper_bounds*(1 + rtol))
        if neg_conc or too_much:
            if neg_conc:
                warnings.warn("Negative concentration")
            if too_much:
                warnings.warn("Too much of at least one component")
            return False
        return True

    def _solve(self, init_concs, x0=None, NumSys=(NumSysLog, NumSysLin),
               neqsys='chained_conditional', **kwargs):
        if isinstance(neqsys, str):
            neqsys = self.get_neqsys(
                neqsys, NumSys=NumSys,
                rref_equil=kwargs.pop('rref_equil', False),
                rref_preserv=kwargs.pop('rref_preserv', False),
                precipitates=kwargs.pop('precipitates', None))
        if x0 is None:
            x0 = init_concs
        params = np.concatenate((init_concs, [float(elem) for elem
                                              in self.eq_constants()]))
        x, sol = neqsys.solve(x0, params, **kwargs)
        if not sol['success']:
            warnings.warn("Root-finding indicated as failed by solver.")
        sane = self._result_is_sane(init_concs, x)
        return x, sol, sane

    def solve(self, init_concs, varied=None, **kwargs):
        results = EqCalcResult(self, init_concs, varied)
        results.solve()
        return results

    def root(self, init_concs, x0=None, neqsys=None, NumSys=NumSysLog,
             neqsys_type='chained_conditional', **kwargs):
        init_concs = self.as_per_substance_array(init_concs)
        params = np.concatenate((init_concs, [float(elem) for elem
                                              in self.eq_constants()]))
        if neqsys is None:
            neqsys = self.get_neqsys(
                neqsys_type, NumSys=NumSys,
                rref_equil=kwargs.pop('rref_equil', False),
                rref_preserv=kwargs.pop('rref_preserv', False),
                precipitates=kwargs.pop('precipitates', None))
        if x0 is None:
            x0 = init_concs
        x, sol = neqsys.solve(x0, params, **kwargs)
        if not sol['success']:
            warnings.warn("Root finding indicated as failed by solver.")
        sane = self._result_is_sane(init_concs, x)
        return x, sol, sane

    @staticmethod
    def _get_default_plot_ax(subplot_kwargs=None):
        import matplotlib.pyplot as plt
        if subplot_kwargs is None:
            subplot_kwargs = dict(xscale='log', yscale='log')
        return plt.subplot(1, 1, 1, **subplot_kwargs)

    def substance_labels(self, latex=False):
        if latex:
            result = ['$' + s.latex_name + '$'
                      for s in self.substances.values()]
            return result
        else:
            return [s.name for s in self.substances.values()]

    def roots(self, init_concs, varied_data, varied, x0=None,
              NumSys=NumSysLog, plot_kwargs=None,
              neqsys_type='chained_conditional', **kwargs):
        """
        Parameters
        ----------
        init_concs : array or dict
        varied_data : array
        varied_idx : int or str
        x0 : array
        NumSys : _NumSys subclass
            See :class:`NumSysLin`, :class:`NumSysLog`, etc.
        plot_kwargs : dict
            See py:meth:`pyneqsys.NeqSys.solve`. Two additional keys
            are intercepted here:
                latex_names: bool (default: False)
                conc_unit_str: str (default: 'M')
        neqsys_type : str
            what method to use for NeqSys construction (get_neqsys_*)
        \\*\\*kwargs :
            Keyword argumetns passed on to py:meth:`pyneqsys.NeqSys.solve_series`.

        """
        _plot = plot_kwargs is not None
        if _plot:
            latex_names = plot_kwargs.pop('latex_names', False)
            conc_unit_str = plot_kwargs.pop('conc_unit_str', 'M')
            if 'ax' not in plot_kwargs:
                plot_kwargs['ax'] = self._get_default_plot_ax()

        init_concs = self.as_per_substance_array(init_concs)
        neqsys = self.get_neqsys(
            neqsys_type, NumSys=NumSys,
            rref_equil=kwargs.pop('rref_equil', False),
            rref_preserv=kwargs.pop('rref_preserv', False),
            precipitates=kwargs.pop('precipitates', None))
        if x0 is None:
            x0 = init_concs

        if _plot:
            cb = neqsys.solve_and_plot_series
            if 'plot_kwargs' not in kwargs:
                kwargs['plot_kwargs'] = plot_kwargs
            if 'labels' not in kwargs['plot_kwargs']:
                kwargs['plot_kwargs']['labels'] = (
                    self.substance_labels(latex_names))
            if 'substances' in plot_kwargs:
                if 'indices' in plot_kwargs:
                    raise ValueError("Now I am confused..")
                kwargs['plot_kwargs']['indices'] = map(
                    self.as_substance_index, plot_kwargs.pop('substances'))
                print(kwargs['plot_kwargs']['indices'])
        else:
            cb = neqsys.solve_series

        params = np.concatenate((init_concs, self.eq_constants()))
        xvecs, info_dicts = cb(
            x0, params, varied_data, self.as_substance_index(varied),
            propagate=False, **kwargs)
        sanity = [self._result_is_sane(init_concs, x) for x in xvecs]

        if _plot:
            import matplotlib.pyplot as plt
            from pyneqsys.plotting import mpl_outside_legend
            mpl_outside_legend(plt.gca())
            varied_subst = self.substances[varied]
            xlbl = ('$[' + varied_subst.latex_name + ']_0$' if latex_names
                    else str(varied_subst))
            plt.gca().set_xlabel(xlbl + ' / ' + conc_unit_str)
            plt.gca().set_ylabel('Concentration / ' + conc_unit_str)

        return xvecs, info_dicts, sanity

    def plot_errors(self, concs, init_concs, varied_data, varied, axes=None,
                    compositions=True, Q=True, subplot_kwargs=None):
        if axes is None:
            import matplotlib.pyplot as plt
            if subplot_kwargs is None:
                subplot_kwargs = dict(xscale='log')
            fig, axes = plt.subplots(1, 2, figsize=(10, 4),
                                     subplot_kw=subplot_kwargs)
        varied_idx = self.as_substance_index(varied)
        ls, c = '- -- : -.'.split(), 'krgbcmy'
        all_inits = np.tile(self.as_per_substance_array(init_concs),
                            (len(varied_data), 1))
        all_inits[:, varied_idx] = varied_data
        if compositions:
            cmp_nrs, m1, m2 = self.composition_conservation(concs, all_inits)
            for cidx, (cmp_nr, a1, a2) in enumerate(zip(cmp_nrs, m1, m2)):
                axes[0].plot(concs[:, varied_idx],
                             a1-a2, label='Comp ' + str(cmp_nr),
                             ls=ls[cidx % len(ls)], c=c[cidx % len(c)])
                axes[1].plot(concs[:, varied_idx],
                             (a1-a2)/np.abs(a2), label='Comp ' + str(cmp_nr),
                             ls=ls[cidx % len(ls)], c=c[cidx % len(c)])

        if Q:
            # TODO: handle precipitate phases in plotting Q error
            qs = self.equilibrium_quotients(concs)
            ks = [rxn.param for rxn in self.rxns]
            for idx, (q, k) in enumerate(zip(qs, ks)):
                axes[0].plot(concs[:, varied_idx],
                             q-k, label='K R:' + str(idx),
                             ls=ls[(idx+cidx) % len(ls)],
                             c=c[(idx+cidx) % len(c)])
                axes[1].plot(concs[:, varied_idx],
                             (q-k)/k, label='K R:' + str(idx),
                             ls=ls[(idx+cidx) % len(ls)],
                             c=c[(idx+cidx) % len(c)])

        from pyneqsys.plotting import mpl_outside_legend
        mpl_outside_legend(axes[0])
        mpl_outside_legend(axes[1])
        axes[0].set_title("Absolute errors")
        axes[1].set_title("Relative errors")



# -*- coding: utf-8 -*-
"""
ChemPy is a Python package useful for solving problems in chemistry.
"""

from __future__ import absolute_import, division, print_function

from ._url import __url__
from ._release import __version__
from .chemistry import Substance, Reaction, Equilibrium, Species, balance_stoichiometry, mass_fractions
from .reactionsystem import ReactionSystem
from .henry import Henry
from .util.periodic import atomic_number
from .kinetics import EyringParam, EyringHS, MassAction

from .util.pyutil import ChemPyDeprecationWarning

import sys

if sys.version_info < (3, 5, 0):
    import warnings
    warnings.warn("Use 'chempy<0.7' if using python versions < 3.5", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .kinetics.arrhenius import (
    arrhenius_equation, fit_arrhenius_equation, ArrheniusParam, ArrheniusParamWithUnits
)


warnings.warn("use .kinetics.arrhenius instead of .arrhenius", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import warnings


class Deprecation(object):
    """ Decorator factory for deprecating functions or classes.

    This class represent deprecations of functions or classes and is designed
    to be used with the ``warnings`` library.

    Parameters
    ----------
    last_supported_version : str, optional
        Version string, e.g. ``'0.2.1'``.
    will_be_missing_in : str, optional
        Version string, e.g. ``'0.3.0'``.
    use_instead : object or str, optional
        Function or class to use instead or descriptive string.
    issue : str, optional
    issues_url : callback, optional
        Converts issue to url, e.g. ``lambda s: 'https://github.com/user/repo/issues/%s/' % s.lstrip('gh-')``.
    warning: DeprecationWarning, optional
        Any subclass of DeprecationWarning, tip: you may invoke:
        ``warnings.simplefilter('once', MyWarning)`` at module init.

    Examples
    --------
    >>> import warnings
    >>> warnings.simplefilter("error", DeprecationWarning)
    >>> @Deprecation()
    ... def f():
    ...     return 1
    ...
    >>> f()  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    DeprecationWarning: f is deprecated.
    >>> @Deprecation(last_supported_version='0.4.0')
    ... def some_old_function(x):
    ...     return x*x - x
    ...
    >>> Deprecation.inspect(some_old_function).last_supported_version
    '0.4.0'
    >>> @Deprecation(will_be_missing_in='1.0')
    ... class ClumsyClass(object):
    ...     pass
    ...
    >>> Deprecation.inspect(ClumsyClass).will_be_missing_in
    '1.0'
    >>> warnings.resetwarnings()

    Notes
    -----
    :class:`DeprecationWarning` is ignored by default. Use custom warning
    and filter appropriately. Alternatively, run python with ``-W`` flag or set
    the appropriate environment variable:

    ::

        $ python -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        $ python -Wd -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        -c:1: DeprecationWarning: X
        $ export PYTHONWARNINGS=d
        $ python -c 'import warnings as w; w.warn("X", DeprecationWarning)'
        -c:1: DeprecationWarning: X

    """

    _deprecations = {}

    def __init__(self, last_supported_version=None, will_be_missing_in=None,
                 use_instead=None, issue=None, issues_url=None,
                 warning=DeprecationWarning):
        if last_supported_version is not None and\
           not isinstance(last_supported_version, (str, tuple, list)) and\
           callable(last_supported_version):
            raise ValueError("last_supported_version not str, tuple or list")

        self.last_supported_version = last_supported_version
        self.will_be_missing_in = will_be_missing_in
        self.use_instead = use_instead
        self.issue = issue
        self.issues_url = issues_url
        self.warning = warning
        self.warning_message = self._warning_message_template()

    @classmethod
    def inspect(cls, obj):
        """ Get the :class:`Deprecation` instance of a deprecated function. """
        return cls._deprecations[obj]

    def _warning_message_template(self):
        msg = '%(func_name)s is deprecated'
        if self.last_supported_version is not None:
            msg += ' since (not including) % s' % self.last_supported_version
        if self.will_be_missing_in is not None:
            msg += ', it will be missing in %s' % self.will_be_missing_in
        if self.issue is not None:
            if self.issues_url is not None:
                msg += self.issues_url(self.issue)
            else:
                msg += ' (see issue %s)' % self.issue
        if self.use_instead is not None:
            try:
                msg += '. Use %s instead' % self.use_instead.__name__
            except AttributeError:
                msg += '. Use %s instead' % self.use_instead
        return msg + '.'

    def __call__(self, wrapped):
        """ Decorates function to be deprecated """
        msg = self.warning_message % {'func_name': wrapped.__name__}
        wrapped_doc = wrapped.__doc__ or ''
        if hasattr(wrapped, '__mro__'):  # wrapped is a class
            class _Wrapper(wrapped):
                __doc__ = msg + '\n\n' + wrapped_doc

                def __init__(_self, *args, **kwargs):
                    warnings.warn(msg, self.warning, stacklevel=2)
                    wrapped.__init__(_self, *args, **kwargs)

        else:  # wrapped is a function
            def _Wrapper(*args, **kwargs):
                warnings.warn(msg, self.warning, stacklevel=2)
                return wrapped(*args, **kwargs)
            _Wrapper.__doc__ = msg + '\n\n' + wrapped_doc

        self._deprecations[_Wrapper] = self
        _Wrapper.__name__ = wrapped.__name__
        _Wrapper.__module__ = wrapped.__module__
        return _Wrapper



# -*- coding: utf-8 -*-
""" The units module provides the following attributes:

- ``chempy.units.default_units``
- ``chempy.units.default_constants``
- ``chempy.units.SI_base_registry``

together with some functions.

Currently `quantities <https://pypi.python.org/pypi/quantities>`_ is used as
the underlying package to handle units. If it is possible you should try to
only use the ``chempy.units`` module (since it is likely that ``ChemPy``
will change this backend at some point in the future). Therefore you should not
rely on any attributes of the ``Quantity`` instances (and rather use
getter & setter functions in `chempy.units`).

"""
from __future__ import (absolute_import, division, print_function)

from functools import reduce
from operator import mul

from .util.arithmeticdict import ArithmeticDict
from .util.pyutil import NameSpace, deprecated

units_library = 'quantities'  # info used for selective testing.


try:
    pq = __import__(units_library)
except ImportError:
    UncertainQuantity = None
    default_constants = None
    default_units = None
    SI_base_registry = None
    np = None
else:
    import numpy as np
    from .util._quantities import _patch_quantities
    _patch_quantities(pq)
    UncertainQuantity = pq.UncertainQuantity
    # Let us extend the underlying pq namespace with some common units in
    # chemistry
    default_constants = pq.constants

    default_units = NameSpace(pq)
    default_units.dm = default_units.decimetre = pq.UnitQuantity(
        'decimetre',  default_units.m / 10.0, u_symbol='dm')
    default_units.m3 = default_units.metre**3
    default_units.dm3 = default_units.decimetre**3
    default_units.cm3 = default_units.centimetre**3
    if not hasattr(default_units, 'molar'):
        default_units.molar = pq.UnitQuantity(
            'M',  1e3 * default_units.mole/default_units.m3,
            u_symbol='M')
    if not hasattr(default_units, 'millimolar'):
        default_units.millimolar = pq.UnitQuantity(
            'mM', 1 * default_units.mole/default_units.m3,
            u_symbol='mM')
    if not hasattr(default_units, 'micromolar'):
        default_units.micromolar = pq.UnitQuantity(
            'uM',  1e-3 * default_units.mole/default_units.m3,
            u_symbol='μM')
    if not hasattr(default_units, 'nanomolar'):
        default_units.nanomolar = pq.UnitQuantity(
            'nM', 1e-6 * default_units.mole/default_units.m3,
            u_symbol='nM')
    if not hasattr(default_units, 'molal'):
        default_units.molal = pq.UnitQuantity(
            'molal',  default_units.mole / default_units.kg,
            u_symbol='molal')
    if not hasattr(default_units, 'per100eV'):
        default_units.per100eV = pq.UnitQuantity(
            'per_100_eV',
            1/(100*default_units.eV*default_constants.Avogadro_constant),
            u_symbol='(100eV)**-1')
    if not hasattr(default_units, 'micromole'):
        default_units.micromole = pq.UnitQuantity(
            'micromole',  pq.mole/1e6,  u_symbol=u'μmol')
    if not hasattr(default_units, 'nanomole'):
        default_units.nanomole = pq.UnitQuantity(
            'nanomole',  pq.mole/1e9,  u_symbol=u'nmol')
    if not hasattr(default_units, 'kilojoule'):
        default_units.kilojoule = pq.UnitQuantity(
            'kilojoule',  1e3*pq.joule,  u_symbol='kJ')
    if not hasattr(default_units, 'kilogray'):
        default_units.kilogray = pq.UnitQuantity(
            'kilogray',  1e3*pq.gray,  u_symbol='kGy')
    if not hasattr(default_units, 'perMolar_perSecond'):
        default_units.perMolar_perSecond = 1/default_units.molar/pq.s
    if not hasattr(default_units, 'umol'):
        default_units.umol = default_units.micromole
    if not hasattr(default_units, 'umol_per_J'):
        default_units.umol_per_J = default_units.umol / pq.joule

    # unit registry data and logic:

    SI_base_registry = {
        'length': default_units.metre,
        'mass': default_units.kilogram,
        'time': default_units.second,
        'current': default_units.ampere,
        'temperature': default_units.kelvin,
        'luminous_intensity': default_units.candela,
        'amount': default_units.mole
    }


def magnitude(value):
    try:
        return value.magnitude
    except AttributeError:
        return value


def simplified(value):
    return value.simplified


def is_quantity(arg):
    if arg.__class__.__name__ == 'Quantity':
        return True  # this checks works even if quantities is not installed.
    else:
        return False


energy = ArithmeticDict(int, {'mass': 1, 'length': 2, 'time': -2})
volume = ArithmeticDict(int, {'length': 3})
concentration = {'amount': 1} - volume


def get_derived_unit(registry, key):
    """ Get the unit of a physcial quantity in a provided unit system.

    Parameters
    ----------
    registry: dict (str: unit)
        mapping 'length', 'mass', 'time', 'current', 'temperature',
        'luminous_intensity', 'amount'. If registry is ``None`` the
        function returns 1.0 unconditionally.
    key: str
        one of the registry keys or one of: 'diffusivity', 'electricalmobility',
        'permittivity', 'charge', 'energy', 'concentration', 'density',
        'radiolyticyield'.

    Examples
    --------
    >>> m, s = default_units.meter, default_units.second
    >>> get_derived_unit(SI_base_registry, 'diffusivity') == m**2/s
    True

    """
    if registry is None:
        return 1.0
    derived = {
        'diffusivity': registry['length']**2/registry['time'],
        'electrical_mobility': (registry['current']*registry['time']**2 /
                                registry['mass']),
        'permittivity': (registry['current']**2*registry['time']**4 /
                         (registry['length']**3*registry['mass'])),
        'charge': registry['current']*registry['time'],
        'energy': registry['mass']*registry['length']**2/registry['time']**2,
        'concentration': registry['amount']/registry['length']**3,
        'density': registry['mass']/registry['length']**3,
    }
    derived['diffusion'] = derived['diffusivity']  # 'diffusion' is deprecated
    derived['radiolytic_yield'] = registry['amount']/derived['energy']
    derived['doserate'] = derived['energy']/registry['mass']/registry['time']
    derived['linear_energy_transfer'] = derived['energy']/registry['length']

    try:
        return derived[key]
    except KeyError:
        return registry[key]


def unit_registry_to_human_readable(unit_registry):
    """ Serialization of a unit registry. """
    if unit_registry is None:
        return None
    new_registry = {}
    for k in SI_base_registry:
        if unit_registry[k] is 1:
            new_registry[k] = 1, 1
        else:
            dim_list = list(unit_registry[k].dimensionality)
            if len(dim_list) != 1:
                raise TypeError("Compound units not allowed: {}".format(
                    dim_list))
            u_symbol = dim_list[0].u_symbol
            new_registry[k] = float(unit_registry[k]), u_symbol
    return new_registry


def _latex_from_dimensionality(dim):
    # see https://github.com/python-quantities/python-quantities/issues/148
    from quantities.markup import format_units_latex
    return format_units_latex(dim, mult=r'\\cdot')


def latex_of_unit(quant):
    """ Returns LaTeX reperesentation of the unit of a quantity

    Examples
    --------
    >>> print(latex_of_unit(1/default_units.kelvin))
    \\mathrm{\\frac{1}{K}}

    """
    return _latex_from_dimensionality(quant.dimensionality).strip('$')


def unicode_of_unit(quant):
    """ Returns unicode reperesentation of the unit of a quantity

    Examples
    --------
    >>> print(unicode_of_unit(1/default_units.kelvin))
    1/K

    """
    return quant.dimensionality.unicode


def html_of_unit(quant):
    """ Returns HTML reperesentation of the unit of a quantity

    Examples
    --------
    >>> print(html_of_unit(2*default_units.m**2))
    m<sup>2</sup>

    """
    return quant.dimensionality.html


def unit_registry_from_human_readable(unit_registry):
    """ Deserialization of unit_registry. """
    if unit_registry is None:
        return None
    new_registry = {}
    for k in SI_base_registry:
        factor, u_symbol = unit_registry[k]
        if u_symbol == 1:
            unit_quants = [1]
        else:
            unit_quants = list(pq.Quantity(0, u_symbol).dimensionality.keys())

        if len(unit_quants) != 1:
            raise TypeError("Unkown UnitQuantity: {}".format(unit_registry[k]))
        else:
            new_registry[k] = factor*unit_quants[0]
    return new_registry


# Abstraction of underlying package providing units and dimensional analysis:

def is_unitless(expr):
    """ Returns ``True`` if ``expr`` is unitless, otherwise ``False``

    Examples
    --------
    >>> is_unitless(42)
    True
    >>> is_unitless(42*default_units.kilogram)
    False

    """
    if hasattr(expr, 'dimensionality'):
        return expr.simplified.dimensionality == pq.dimensionless.dimensionality
    if isinstance(expr, dict):
        return all(is_unitless(_) for _ in expr.values())
    elif isinstance(expr, (tuple, list)):
        return all(is_unitless(_) for _ in expr)
    return True


def unit_of(expr, simplified=False):
    """ Returns the unit of a quantity

    Examples
    --------
    >>> unit_of(42*pq.second) == unit_of(12*pq.second)
    True
    >>> unit_of(42)
    1

    """
    if isinstance(expr, (tuple, list)):
        return unit_of(uniform(expr)[0], simplified)
    elif isinstance(expr, dict):
        return unit_of(list(uniform(expr).values())[0], simplified)

    try:
        if simplified:
            return expr.units.simplified
        else:
            return expr.units
    except AttributeError:
        return 1


def rescale(value, unit):
    try:
        return value.rescale(unit)
    except AttributeError:
        if unit == 1:
            return value
        else:
            raise


def to_unitless(value, new_unit=None):
    """ Nondimensionalization of a quantity.

    Parameters
    ----------
    value: quantity
    new_unit: unit

    Examples
    --------
    >>> '%.1g' % to_unitless(1*default_units.metre, default_units.nm)
    '1e+09'
    >>> '%.1g %.1g' % tuple(to_unitless([1*default_units.m, 1*default_units.mm], default_units.nm))
    '1e+09 1e+06'

    """
    if new_unit is None:
        new_unit = pq.dimensionless

    if isinstance(value, (list, tuple)):
        return np.array([to_unitless(elem, new_unit) for elem in value])
    elif isinstance(value, np.ndarray) and not hasattr(value, 'rescale'):
        if new_unit == 1 and value.dtype != object:
            return value
        return np.array([to_unitless(elem, new_unit) for elem in value])
    elif isinstance(value, dict):
        new_value = dict(value.items())  # value.copy()
        for k in value:
            new_value[k] = to_unitless(value[k], new_unit)
        return new_value
    elif isinstance(value, (int, float)) and new_unit is 1 or new_unit is None:
        return value
    elif isinstance(value, str):
        raise ValueError("str not supported")
    else:
        try:
            try:
                result = (value*pq.dimensionless/new_unit).rescale(pq.dimensionless)
            except AttributeError:
                if new_unit == pq.dimensionless:
                    return value
                else:
                    raise
            else:
                if result.ndim == 0:
                    return float(result)
                else:
                    return np.asarray(result)
        except TypeError:
            return np.array([to_unitless(elem, new_unit) for elem in value])


def uniform(container):
    """ Turns a list, tuple or dict with mixed units into one with uniform units.

    Parameters
    ----------
    container : tuple, list or dict

    Examples
    --------
    >>> km, m = default_units.kilometre, default_units.metre
    >>> uniform(dict(a=3*km, b=200*m))  # doctest: +SKIP
    {'b': array(200.0) * m, 'a': array(3000.0) * m}

    """
    if isinstance(container, (tuple, list)):
        unit = unit_of(container[0])
    elif isinstance(container, dict):
        unit = unit_of(list(container.values())[0])
        return container.__class__([(k, to_unitless(v, unit)*unit) for k, v in container.items()])
    else:
        return container
    return to_unitless(container, unit)*unit


def get_physical_dimensionality(value):
    if is_unitless(value):
        return {}
    _quantities_mapping = {
        pq.UnitLength: 'length',
        pq.UnitMass: 'mass',
        pq.UnitTime: 'time',
        pq.UnitCurrent: 'current',
        pq.UnitTemperature: 'temperature',
        pq.UnitLuminousIntensity: 'luminous_intensity',
        pq.UnitSubstance: 'amount'
    }
    return {_quantities_mapping[k.__class__]: v for k, v
            in uniform(value).simplified.dimensionality.items()}

get_physical_quantity = deprecated(
    use_instead=get_physical_dimensionality,
    will_be_missing_in='0.8.0'
)(get_physical_dimensionality)


def _get_unit_from_registry(dimensionality, registry):
    return reduce(mul, [registry[k]**v for k, v in dimensionality.items()])


def default_unit_in_registry(value, registry):
    _dimensionality = get_physical_dimensionality(value)
    if _dimensionality == {}:
        return 1
    return _get_unit_from_registry(_dimensionality, registry)


def unitless_in_registry(value, registry):
    _default_unit = default_unit_in_registry(value, registry)
    return to_unitless(value, _default_unit)


# NumPy like functions for compatibility:

def compare_equality(a, b):
    """ Returns True if two arguments are equal.

    Both arguments need to have the same dimensionality.

    Parameters
    ----------
    a : quantity
    b : quantity

    Examples
    --------
    >>> km, m = default_units.kilometre, default_units.metre
    >>> compare_equality(3*km, 3)
    False
    >>> compare_equality(3*km, 3000*m)
    True

    """
    # Work around for https://github.com/python-quantities/python-quantities/issues/146
    try:
        a + b
    except TypeError:
        # We might be dealing with e.g. None (None + None raises TypeError)
        try:
            len(a)
        except TypeError:
            # Assumed scalar
            return a == b
        else:
            if len(a) != len(b):
                return False
            return all(compare_equality(_a, _b) for _a, _b in zip(a, b))
    except ValueError:
        return False
    else:
        return a == b


def allclose(a, b, rtol=1e-8, atol=None):
    """ Analogous to ``numpy.allclose``. """
    try:
        d = abs(a - b)
    except Exception:
        try:
            if len(a) == len(b):
                return all(allclose(_a, _b, rtol, atol) for _a, _b in zip(a, b))
            else:
                return False
        except Exception:
            return False
    lim = abs(a)*rtol
    if atol is not None:
        lim += atol

    try:
        len(d)
    except TypeError:
        return d <= lim
    else:
        try:
            len(lim)
        except TypeError:
            return np.all([_d <= lim for _d in d])
        else:
            return np.all([_d <= _lim for _d, _lim in zip(d, lim)])


def linspace(start, stop, num=50):
    """ Analogous to ``numpy.linspace``.

    Examples
    --------
    >>> abs(linspace(2, 8, num=3)[1] - 5) < 1e-15
    True

    """

    # work around for quantities v0.10.1 and NumPy
    unit = unit_of(start)
    start_ = to_unitless(start, unit)
    stop_ = to_unitless(stop, unit)
    return np.linspace(start_, stop_, num)*unit


def logspace_from_lin(start, stop, num=50):
    """ Logarithmically spaced data points

    Examples
    --------
    >>> abs(logspace_from_lin(2, 8, num=3)[1] - 4) < 1e-15
    True

    """
    unit = unit_of(start)
    start_ = np.log2(to_unitless(start, unit))
    stop_ = np.log2(to_unitless(stop, unit))
    return np.exp2(np.linspace(start_, stop_, num))*unit


def _sum(iterable):
    try:
        result = next(iterable)
    except TypeError:
        result = iterable[0]
        for elem in iterable[1:]:
            result += elem
        return result
    else:
        try:
            while True:
                result += next(iterable)
        except StopIteration:
            return result
        else:
            raise ValueError("Not sure how this point was reached")


class Backend(object):
    """ Wrapper around modules such as numpy and math

    Instances of Backend wraps a module, e.g. `numpy` and ensures that
    arguments passed on are unitless, i.e. it raises an error if a
    transcendental function is used with quantities with units.

    Parameters
    ----------
    underlying_backend : module, str or tuple of str
        e.g. 'numpy' or ('sympy', 'math')

    Examples
    --------
    >>> import math
    >>> km, m = default_units.kilometre, default_units.metre
    >>> math.exp(3*km) == math.exp(3*m)
    True
    >>> be = Backend('math')
    >>> be.exp(3*km)
    Traceback (most recent call last):
        ...
    ValueError: Unable to convert between units of "km" and "dimensionless"
    >>> import numpy as np
    >>> np.sum([1000*pq.metre/pq.kilometre, 1])
    1001.0
    >>> be_np = Backend(np)
    >>> be_np.sum([[1000*pq.metre/pq.kilometre, 1], [3, 4]], axis=1)
    array([2., 7.])

    """

    def __init__(self, underlying_backend=('numpy', 'math')):
        if isinstance(underlying_backend, tuple):
            for name in underlying_backend:
                try:
                    self.be = __import__(name)
                except ImportError:
                    continue
                else:
                    break
            else:
                raise ValueError("Could not import any of %s" %
                                 str(underlying_backend))
        elif isinstance(underlying_backend, str):
            self.be = __import__(underlying_backend)
        else:
            self.be = underlying_backend

    def __getattr__(self, attr):
        be_attr = getattr(self.be, attr)
        if callable(be_attr):
            return lambda *args, **kwargs: be_attr(*map(to_unitless, args), **kwargs)
        else:
            return be_attr


# TODO: decide whether to deprecate in favor of "number_to_scientific_latex"?
def format_string(value, precision='%.5g', tex=False):
    """ Formats a scalar with unit as two strings

    Parameters
    ----------
    value: float with unit
    precision: str
    tex: bool
       LaTeX formatted or not? (no '$' signs)

    Examples
    --------
    >>> print(' '.join(format_string(0.42*default_units.mol/default_units.decimetre**3)))
    0.42 mol/decimetre**3
    >>> print(' '.join(format_string(2/default_units.s, tex=True)))
    2 \\mathrm{\\frac{1}{s}}

    """
    if tex:
        unit_str = latex_of_unit(value)
    else:
        from quantities.markup import config
        attr = 'unicode' if config.use_unicode else 'string'
        unit_str = getattr(value.dimensionality, attr)
    return precision % float(value.magnitude), unit_str


def concatenate(arrays, **kwargs):
    """ Patched version of numpy.concatenate

    Examples
    --------
    >>> from chempy.units import default_units as u
    >>> all(concatenate(([2, 3]*u.s, [4, 5]*u.s)) == [2, 3, 4, 5]*u.s)
    True

    """
    unit = unit_of(arrays[0])
    result = np.concatenate([to_unitless(arr, unit) for arr in arrays], **kwargs)
    return result*unit


def tile(array, *args, **kwargs):
    """ Parched version of numpy.tile """
    try:
        elem = array[0, ...]
    except TypeError:
        elem = array[0]

    unit = unit_of(elem)
    result = np.tile(to_unitless(array, unit), *args, **kwargs)
    return result*unit


def polyfit(x, y, deg, **kwargs):
    u_x = unit_of(x[0])
    u_y = unit_of(y[0])
    _x, _y = to_unitless(x, u_x), to_unitless(y, u_y)
    p = np.polyfit(_x, _y, deg)
    return [v*u_y*u_x**(i-deg) for i, v in enumerate(p)]


def polyval(p, x):
    try:
        u_x = unit_of(x[0])
    except (TypeError, IndexError):
        u_x = unit_of(x)
    u_y = unit_of(p[-1])
    deg = len(p) - 1
    _p = [to_unitless(v, u_y*u_x**(i-deg)) for i, v in enumerate(p)]
    _x = to_unitless(x, u_x)
    _y = np.polyval(_p, _x)
    return _y*u_y

patched_numpy = NameSpace(np)
patched_numpy.allclose = allclose
patched_numpy.concatenate = concatenate
patched_numpy.linspace = linspace
patched_numpy.tile = tile
patched_numpy.polyfit = polyfit
patched_numpy.polyval = polyval



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from math import log as ln

from ..electrolytes import A as A_dh, B as B_dh
from ..electrolytes import limiting_log_gamma, _ActivityProductBase, ionic_strength
from ..units import (allclose, units_library,
                     default_constants as consts,
                     default_units as u)
from ..util.testing import requires


def test_ionic_strength():
    assert abs(ionic_strength([0.1, 1.3, 2.1, 0.7],
                              [-1, 2, -3, 4], warn=False) - 17.7) < 1e-14


def test_ActivityProductBase():
    ap = _ActivityProductBase((1, -2, 3), 17, 42)
    assert ap.stoich == (1, -2, 3)
    assert ap.args == (17, 42)


def test_A():
    A20 = A_dh(80.1, 293.15, 998.2071)/ln(10)
    assert abs(A20 - 0.50669) < 1e-5


@requires(units_library)
def test_A__units():
    A20q = A_dh(80.1, 293.15*u.K, 998.2071*u.kg/u.m**3,
                b0=1*u.mol/u.kg, constants=consts)
    assert abs(A20q - 0.50669) < 1e-5


def test_B():
    assert abs(1e-10*B_dh(80.1, 293.15, 998.2071, 1) - 0.3282) < 1e-3


@requires(units_library)
def test_B__units():
    B20q = B_dh(80.1, 293.15*u.K, 998.2071*u.kg/u.m**3,
                b0=u.mol/u.kg, constants=consts, units=u)
    close = allclose(B20q.simplified, 0.3282/u.angstrom, rtol=1e-3)
    assert close


def test_limiting_log_gamma():
    A20 = A_dh(80.1, 293.15, 998.2071)/ln(10)
    log_gamma = limiting_log_gamma(0.4, -3, A20)
    assert abs(log_gamma + 2.884130) < 1e-4



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import collections

import pytest
try:
    import numpy as np
except ImportError:
    np, NumSysLin, NumSysLog = [None]*3
else:
    from ..equilibria import EqSystem, NumSysLin, NumSysLog
    from .ammonical_cupric_solution import get_ammonical_cupric_eqsys

from ..util.testing import requires
from ..chemistry import (
    Substance, Reaction, Equilibrium, Species
)


@requires('numpy')
def test_EqSystem():
    a, b = sbstncs = Substance('a'), Substance('b')
    rxns = [Reaction({'a': 1}, {'b': 1})]
    es = EqSystem(rxns, collections.OrderedDict(
        [(s.name, s) for s in sbstncs]))
    assert es.net_stoichs().tolist() == [[-1, 1]]


def _get_es1():
    a, b = sbstncs = Species('a'), Species('b')
    rxns = [Equilibrium({'a': 1}, {'b': 1}, 3)]
    return EqSystem(rxns, sbstncs)


def _get_es_water(EqSys=None):
    if EqSys is None:
        EqSys = EqSystem
    H2O = Substance('H2O', charge=0, composition={1: 2, 8: 1})
    OHm = Substance('OH-', charge=-1, composition={1: 1, 8: 1})
    Hp = Substance('H+', charge=1, composition={1: 1})
    Kw = 1e-14/55.5
    w_auto_p = Equilibrium({'H2O': 1}, {'Hp': 1, 'OHm': 1}, Kw)
    return EqSys([w_auto_p], [H2O, OHm, Hp])


@requires('numpy')
def test_EqSystem_1():
    es = _get_es1()
    assert es.stoichs().tolist() == [[-1, 1]]
    assert es.eq_constants() == [3]


@requires('numpy')
def test_Equilibria_arithmetics():
    es1 = _get_es1()
    e, = es1.rxns
    e2 = 2*e
    sum2 = e + e
    assert sum2 == e2


@pytest.mark.slow
@requires('numpy')
def test_Equilibria_root():
    eqsys, c0 = get_ammonical_cupric_eqsys()
    x, sol, sane = eqsys.root(c0, NumSys=(NumSysLog,))
    assert sol['success'] and sane


def _species(Cls):
    return (
        Cls('H2O', 0, composition={1: 2, 8: 1}),
        Cls('H+', 1, composition={1: 1}),
        Cls('OH-', -1, composition={1: 1, 8: 1}),
        Cls('NH4+', 1, composition={1: 4, 7: 1}),
        Cls('NH3', 0, composition={1: 3, 7: 1})
    )


@requires('numpy')
def test_Equilibria_root_simple():
    species = water, hydronium, hydroxide, ammonium, ammonia = _species(Species)

    water_auto_protolysis = Equilibrium(
        {water.name: 1}, {hydronium.name: 1, hydroxide.name: 1}, 1e-14/55.5)
    ammonia_protolysis = Equilibrium(
        {ammonium.name: 1}, {hydronium.name: 1, ammonia.name: 1},
        10**-9.26/55.5
    )
    eqsys = EqSystem([water_auto_protolysis, ammonia_protolysis], species)

    init_concs = collections.defaultdict(float, {
        water.name: 55.5, ammonia.name: 1e-3})
    x, sol1, sane1 = eqsys.root(init_concs)
    x, sol2, sane2 = eqsys.root(init_concs, x, NumSys=(NumSysLog,))
    assert sane2
    ref = eqsys.as_per_substance_array({
        water.name: 55.5,
        ammonia.name: 1e-3 - 6.2e-4,
        ammonium.name: 6.2e-4,
        hydronium.name: 1.6e-11,
        hydroxide.name: 6.2e-4
    })
    assert np.allclose(x, ref, rtol=0.02, atol=1e-16)


def _get_NaCl(Cls, **precipitate_kwargs):
    Na_p, Cl_m, NaCl = sbstncs = (
        Cls('Na+', 1, composition={11: 1}),
        Cls('Cl-', -1, composition={17: 1}),
        Cls('NaCl', composition={11: 1, 17: 1}, **precipitate_kwargs)
    )
    sp = Equilibrium({'NaCl': 1}, {'Na+': 1, 'Cl-': 1}, 4.0)
    eqsys = EqSystem([sp], sbstncs)
    cases = [
        [(0, 0, .1), (.1, .1, 0)],
        [(.5, .5, .4), (.9, .9, 0)],
        [(1, 1, 1), (2, 2, 0)],
        [(0, 0, 2), (2, 2, 0)],
        [(3, 3, 3), (2, 2, 4)],
        [(3, 3, 0), (2, 2, 1)],
        [(0, 0, 3), (2, 2, 1)],
    ]
    return eqsys, [s.name for s in sbstncs], cases


@requires('numpy')
def test_EqSystem_dissolved():
    eqsys, names, _ = _get_NaCl(Cls=Species, phase_idx=1)
    inp = eqsys.as_per_substance_array({'Na+': 1, 'Cl-': 2, 'NaCl': 4})
    result = eqsys.dissolved(inp)
    ref = eqsys.as_per_substance_array({'Na+': 5, 'Cl-': 6, 'NaCl': 0})
    assert np.allclose(result, ref)


@requires('numpy')
@pytest.mark.parametrize('NumSys', [(NumSysLin,), (NumSysLog,),
                                    (NumSysLog, NumSysLin)])
def test_precipitate(NumSys):
    eqsys, species, cases = _get_NaCl(Species, phase_idx=1)

    for init, final in cases:
        x, sol, sane = eqsys.root(dict(zip(species, init)),
                                  NumSys=NumSys, rref_preserv=True, tol=1e-12)
        assert sol['success'] and sane
        assert x is not None
        assert np.allclose(x, np.asarray(final))



"""
This package collects functions useful for studying chemical kinetics problems.
"""
from __future__ import (absolute_import, division, print_function)

from .rates import MassAction, EyringHS
from .eyring import EyringParam
from .arrhenius import ArrheniusParam
